import warnings

from ludwig.utils.logging_utils import log_once


def warn_structure_refactor(old_module: str, new_module: str, direct: bool = True) -> None:
    """Create structure refactor warning to indicate modules new location post.

    Only creates a warning once per module.
    """
    old_module = old_module.replace(".py", "")
    if log_once(old_module):
        warning = (
            f"The module `{old_module}` has been moved to `{new_module}` and the old "
            f"location will be deprecated soon. Please adjust your imports to point "
            f"to the new location."
        )

        if direct:
            warning += f" Example: Do a global search and " f"replace `{old_module}` with `{new_module}`."
        else:
            warning += (
                f"\nATTENTION: This module may have been split or refactored. Please "
                f"check the contents of `{new_module}` before making changes."
            )

        with warnings.catch_warnings():
            warnings.simplefilter("always")
            warnings.warn(warning, DeprecationWarning, stacklevel=3)



#! /usr/bin/env python
#
# Checks all dataset download links (just those with URLs, not including kaggle datasets)."""
#
import logging

import requests

import ludwig.datasets

logger = logging.getLogger(__name__)


def check_link(url):
    response = requests.get(url)
    if response.status_code == 200:
        logger.info(f"OK: {url}")
        return True
    else:
        logger.info(f"{response.status_code}: {url}")
        logger.error(response)
        return False


def test_dead_links():
    # Iterate through all datasets, ensure links are valid and reachable.
    all_datasets = ludwig.datasets.list_datasets()

    for dataset_name in all_datasets:
        config = ludwig.datasets.get_dataset_config(dataset_name)
        download_urls = [config.download_urls] if isinstance(config.download_urls, str) else config.download_urls
        for url in download_urls:
            try:
                check_link(url)
            except Exception as e:
                logger.exception(f"Exception thrown downloading {dataset_name} dataset", e)


if __name__ == "__main__":
    test_dead_links()



from typing import Any, Dict

# The dictionary representation of a whylogs.core.view.column_profile_view.ColumnProfileView
# Obtained by calling ColumnProfileView.to_summary_dict().
# See profiling/contants for a list of keys.
ColumnProfileSummary = Dict[str, Any]



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import pandas as pd

from ludwig.backend import LOCAL_BACKEND
from ludwig.constants import COLUMN, NAME, PROC_COLUMN
from ludwig.features.feature_utils import compute_feature_hash
from ludwig.features.number_feature import NumberFeatureMixin


def number_feature():
    feature = {NAME: "x", COLUMN: "x", "type": "number"}
    feature[PROC_COLUMN] = compute_feature_hash(feature)
    return feature


data_df = pd.DataFrame(pd.Series([2, 4, 6, 8, 10]), columns=["x"])

proc_df = pd.DataFrame(columns=["x"])


def test_norm():
    feature_1_meta = NumberFeatureMixin.get_feature_meta(data_df["x"], {"normalization": "zscore"}, LOCAL_BACKEND, True)
    feature_2_meta = NumberFeatureMixin.get_feature_meta(data_df["x"], {"normalization": "minmax"}, LOCAL_BACKEND, True)
    feature_3_meta = NumberFeatureMixin.get_feature_meta(data_df["x"], {"normalization": "iq"}, LOCAL_BACKEND, True)

    assert feature_1_meta["mean"] == 6
    assert feature_2_meta["min"] == 2
    assert feature_2_meta["max"] == 10
    assert feature_3_meta["q1"] == 4
    assert feature_3_meta["q2"] == 6
    assert feature_3_meta["q3"] == 8

    # value checks after normalization
    num_feature = number_feature()

    NumberFeatureMixin.add_feature_data(
        feature_config=num_feature,
        input_df=data_df,
        proc_df=proc_df,
        metadata={num_feature[NAME]: feature_1_meta},
        preprocessing_parameters={"normalization": "zscore"},
        backend=LOCAL_BACKEND,
        skip_save_processed_input=False,
    )
    assert np.allclose(
        np.array(proc_df[num_feature[PROC_COLUMN]]), np.array([-1.26491106, -0.63245553, 0, 0.63245553, 1.26491106])
    )

    NumberFeatureMixin.add_feature_data(
        feature_config=num_feature,
        input_df=data_df,
        proc_df=proc_df,
        metadata={num_feature[NAME]: feature_2_meta},
        preprocessing_parameters={"normalization": "minmax"},
        backend=LOCAL_BACKEND,
        skip_save_processed_input=False,
    )
    assert np.allclose(np.array(proc_df[num_feature[PROC_COLUMN]]), np.array([0, 0.25, 0.5, 0.75, 1]))

    NumberFeatureMixin.add_feature_data(
        feature_config=num_feature,
        input_df=data_df,
        proc_df=proc_df,
        metadata={num_feature[NAME]: feature_3_meta},
        preprocessing_parameters={"normalization": "iq"},
        backend=LOCAL_BACKEND,
        skip_save_processed_input=False,
    )
    assert np.allclose(np.array(proc_df[num_feature[PROC_COLUMN]]), np.array([-1, -0.5, 0, 0.5, 1]))



from typing import List

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import DATE
from ludwig.schema import utils as schema_utils
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import register_encoder_config
from ludwig.schema.metadata import ENCODER_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@register_encoder_config("embed", DATE)
@ludwig_dataclass
class DateEmbedConfig(BaseEncoderConfig):
    @staticmethod
    def module_name():
        return "DateEmbed"

    type: str = schema_utils.ProtectedString(
        "embed",
        description=ENCODER_METADATA["DateEmbed"]["type"].long_description,
    )

    dropout: float = schema_utils.FloatRange(
        default=0.0,
        min=0,
        max=1,
        description="Dropout probability for the embedding.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["dropout"],
    )

    activation: str = schema_utils.ActivationOptions(
        description="The default activation function that will be used for each layer.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["activation"],
    )

    use_bias: bool = schema_utils.Boolean(
        default=True,
        description="Whether the layer uses a bias vector.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["use_bias"],
    )

    bias_initializer: str = schema_utils.InitializerOptions(
        default="zeros",
        description="Initializer to use for the bias vector.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["bias_initializer"],
    )

    weights_initializer: str = schema_utils.InitializerOptions(
        description="Initializer to use for the weights matrix.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["weights_initializer"],
    )

    embedding_size: int = schema_utils.PositiveInteger(
        default=10,
        description="The maximum embedding size adopted.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["embedding_size"],
    )

    embeddings_on_cpu: bool = schema_utils.Boolean(
        default=False,
        description="Whether to force the placement of the embedding matrix in regular memory and have the CPU "
        "resolve them.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["embeddings_on_cpu"],
    )

    output_size: int = schema_utils.PositiveInteger(
        default=10,
        description="If an output_size is not already specified in fc_layers this is the default output_size that "
        "will be used for each layer. It indicates the size of the output of a fully connected layer.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["output_size"],
    )

    norm: str = schema_utils.StringOptions(
        ["batch", "layer"],
        default=None,
        allow_none=True,
        description="The default norm that will be used for each layer.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["norm"],
    )

    norm_params: dict = schema_utils.Dict(
        default=None,
        description="Parameters used if norm is either `batch` or `layer`.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["norm_params"],
    )

    num_fc_layers: int = schema_utils.NonNegativeInteger(
        default=0,
        description="The number of stacked fully connected layers.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["num_fc_layers"],
    )

    # TODO (Connor): Add nesting logic for fc_layers, see fully_connected_module.py
    fc_layers: List[dict] = schema_utils.DictList(
        default=None,
        description="List of dictionaries containing the parameters for each fully connected layer.",
        parameter_metadata=ENCODER_METADATA["DateEmbed"]["fc_layers"],
    )


@DeveloperAPI
@register_encoder_config("wave", DATE)
@ludwig_dataclass
class DateWaveConfig(BaseEncoderConfig):
    @staticmethod
    def module_name():
        return "DateWave"

    type: str = schema_utils.ProtectedString(
        "wave",
        description=ENCODER_METADATA["DateWave"]["type"].long_description,
    )

    dropout: float = schema_utils.FloatRange(
        default=0.0,
        min=0,
        max=1,
        description="Dropout probability for the embedding.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["dropout"],
    )

    activation: str = schema_utils.ActivationOptions(
        description="The default activation function that will be used for each layer.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["activation"],
    )

    use_bias: bool = schema_utils.Boolean(
        default=True,
        description="Whether the layer uses a bias vector.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["use_bias"],
    )

    bias_initializer: str = schema_utils.InitializerOptions(
        default="zeros",
        description="Initializer to use for the bias vector.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["bias_initializer"],
    )

    weights_initializer: str = schema_utils.InitializerOptions(
        description="Initializer to use for the weights matrix.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["weights_initializer"],
    )

    output_size: int = schema_utils.PositiveInteger(
        default=10,
        description="If an output_size is not already specified in fc_layers this is the default output_size that "
        "will be used for each layer. It indicates the size of the output of a fully connected layer.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["output_size"],
    )

    norm: str = schema_utils.StringOptions(
        ["batch", "layer"],
        default=None,
        allow_none=True,
        description="The default norm that will be used for each layer.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["norm"],
    )

    norm_params: dict = schema_utils.Dict(
        default=None,
        description="Parameters used if norm is either `batch` or `layer`.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["norm_params"],
    )

    num_fc_layers: int = schema_utils.PositiveInteger(
        default=1,
        description="The number of stacked fully connected layers.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["num_fc_layers"],
    )

    # TODO (Connor): Add nesting logic for fc_layers, see fully_connected_module.py
    fc_layers: List[dict] = schema_utils.DictList(
        default=None,
        description="List of dictionaries containing the parameters for each fully connected layer.",
        parameter_metadata=ENCODER_METADATA["DateWave"]["fc_layers"],
    )



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC, abstractmethod

from torch import nn

from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.torch_utils import LudwigModule


@DeveloperAPI
class Encoder(LudwigModule, ABC):
    @abstractmethod
    def forward(self, inputs, training=None, mask=None):
        raise NotImplementedError

    def get_embedding_layer(self) -> nn.Module:
        """Returns layer that embeds inputs, used for computing explanations."""
        return next(self.children())

    @property
    def name(self):
        return self.__class__.__name__



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import io
import json
import logging
import os
import sys
import tempfile

import pandas as pd
import torch
from torchvision.io import decode_image

from ludwig.api import LudwigModel
from ludwig.constants import AUDIO, COLUMN
from ludwig.contrib import add_contrib_callback_args
from ludwig.globals import LUDWIG_VERSION
from ludwig.utils.print_utils import get_logging_level_registry, print_ludwig
from ludwig.utils.server_utils import NumpyJSONResponse

logger = logging.getLogger(__name__)

try:
    import uvicorn
    from fastapi import FastAPI
    from starlette.datastructures import UploadFile
    from starlette.middleware import Middleware
    from starlette.middleware.cors import CORSMiddleware
    from starlette.requests import Request
except ImportError as e:
    logger.error(e)
    logger.error(
        " fastapi and other serving dependencies cannot be loaded"
        "and may have not been installed. "
        "In order to install all serving dependencies run "
        "pip install ludwig[serve]"
    )
    sys.exit(-1)

ALL_FEATURES_PRESENT_ERROR = {"error": "entry must contain all input features"}

COULD_NOT_RUN_INFERENCE_ERROR = {"error": "Unexpected Error: could not run inference on model"}


def server(model, allowed_origins=None):
    middleware = [Middleware(CORSMiddleware, allow_origins=allowed_origins)] if allowed_origins else None
    app = FastAPI(middleware=middleware)

    config = model.config
    input_features = {f[COLUMN] for f in config["input_features"]}

    @app.get("/")
    def check_health():
        return NumpyJSONResponse({"message": "Ludwig server is up"})

    @app.post("/predict")
    async def predict(request: Request):
        try:
            form = await request.form()
            entry, files = convert_input(form, model.model.input_features)
        except Exception:
            logger.exception("Failed to parse predict form")
            return NumpyJSONResponse(COULD_NOT_RUN_INFERENCE_ERROR, status_code=500)

        try:
            if (entry.keys() & input_features) != input_features:
                missing_features = set(input_features) - set(entry.keys())
                return NumpyJSONResponse(
                    {
                        "error": "Data received does not contain all input features. "
                        f"Missing features: {missing_features}."
                    },
                    status_code=400,
                )
            try:
                resp, _ = model.predict(dataset=[entry], data_format=dict)
                resp = resp.to_dict("records")[0]
                return NumpyJSONResponse(resp)
            except Exception as exc:
                logger.exception(f"Failed to run predict: {exc}")
                return NumpyJSONResponse(COULD_NOT_RUN_INFERENCE_ERROR, status_code=500)
        finally:
            for f in files:
                os.remove(f.name)

    @app.post("/batch_predict")
    async def batch_predict(request: Request):
        try:
            form = await request.form()
            data, files = convert_batch_input(form, model.model.input_features)
            data_df = pd.DataFrame.from_records(data["data"], index=data.get("index"), columns=data["columns"])
        except Exception:
            logger.exception("Failed to parse batch_predict form")
            return NumpyJSONResponse(COULD_NOT_RUN_INFERENCE_ERROR, status_code=500)

        if (set(data_df.columns) & input_features) != input_features:
            missing_features = set(input_features) - set(data_df.columns)
            return NumpyJSONResponse(
                {
                    "error": "Data received does not contain all input features. "
                    f"Missing features: {missing_features}."
                },
                status_code=400,
            )
        try:
            resp, _ = model.predict(dataset=data_df)
            resp = resp.to_dict("split")
            return NumpyJSONResponse(resp)
        except Exception:
            logger.exception("Failed to run batch_predict: {}")
            return NumpyJSONResponse(COULD_NOT_RUN_INFERENCE_ERROR, status_code=500)

    return app


def _write_file(v, files):
    # Convert UploadFile to a NamedTemporaryFile to ensure it's on the disk
    suffix = os.path.splitext(v.filename)[1]
    named_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
    files.append(named_file)
    named_file.write(v.file.read())
    named_file.close()
    return named_file.name


def _read_image_buffer(v):
    # read bytes sent via REST API and convert to image tensor
    # in [channels, height, width] format
    byte_string = io.BytesIO(v.file.read()).read()
    image = decode_image(torch.frombuffer(byte_string, dtype=torch.uint8))
    return image  # channels, height, width


def convert_input(form, input_features):
    """Returns a new input and a list of files to be cleaned up."""
    new_input = {}
    files = []
    for k, v in form.multi_items():
        if type(v) == UploadFile:
            # check if audio or image file
            if input_features.get(k).type() == AUDIO:
                new_input[k] = _write_file(v, files)
            else:
                new_input[k] = _read_image_buffer(v)
        else:
            new_input[k] = v

    return new_input, files


def convert_batch_input(form, input_features):
    """Returns a new input and a list of files to be cleaned up."""
    file_index = {}
    files = []
    for k, v in form.multi_items():
        if type(v) == UploadFile:
            file_index[v.filename] = v

    data = json.loads(form["dataset"])
    for row in data["data"]:
        for i in range(len(row)):
            if row[i] in file_index:
                feature_name = data["columns"][i]
                if input_features.get(feature_name).type() == AUDIO:
                    row[i] = _write_file(file_index[row[i]], files)
                else:
                    row[i] = _read_image_buffer(file_index[row[i]])

    return data, files


def run_server(
    model_path: str,
    host: str,
    port: int,
    allowed_origins: list,
) -> None:
    """Loads a pre-trained model and serve it on an http server.

    # Inputs

    :param model_path: (str) filepath to pre-trained model.
    :param host: (str, default: `0.0.0.0`) host ip address for the server to use.
    :param port: (int, default: `8000`) port number for the server to use.
    :param allowed_origins: (list) list of origins allowed to make cross-origin requests.

    # Return

    :return: (`None`)
    """
    # Use local backend for serving to use pandas DataFrames.
    model = LudwigModel.load(model_path, backend="local")
    app = server(model, allowed_origins)
    uvicorn.run(app, host=host, port=port)


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script serves a pretrained model", prog="ludwig serve", usage="%(prog)s [options]"
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)

    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    # ----------------
    # Server parameters
    # ----------------
    parser.add_argument(
        "-p",
        "--port",
        help="port for server (default: 8000)",
        default=8000,
        type=int,
    )

    parser.add_argument("-H", "--host", help="host for server (default: 0.0.0.0)", default="0.0.0.0")

    parser.add_argument(
        "-ao",
        "--allowed_origins",
        nargs="*",
        help="A list of origins that should be permitted to make cross-origin requests. "
        'Use "*" to allow any origin. See https://www.starlette.io/middleware/#corsmiddleware.',
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("serve", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.serve")

    print_ludwig("Serve", LUDWIG_VERSION)

    run_server(args.model_path, args.host, args.port, args.allowed_origins)


if __name__ == "__main__":
    cli(sys.argv[1:])



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Union

import numpy as np
import pandas as pd
import torch
from torch import nn

from ludwig.constants import COLUMN, HIDDEN, LOGITS, NAME, NUMBER, PREDICTIONS, PROC_COLUMN
from ludwig.features.base_feature import BaseFeatureMixin, InputFeature, OutputFeature, PredictModule
from ludwig.schema.features.number_feature import NumberInputFeatureConfig, NumberOutputFeatureConfig
from ludwig.types import (
    FeatureMetadataDict,
    FeaturePostProcessingOutputDict,
    PreprocessingConfigDict,
    TrainingSetMetadataDict,
)
from ludwig.utils import output_feature_utils
from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.types import TorchscriptPreprocessingInput

logger = logging.getLogger(__name__)


class NumberTransformer(nn.Module, ABC):
    @abstractmethod
    def transform(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def inverse_transform(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        pass

    @abstractmethod
    def inverse_transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        pass

    @staticmethod
    @abstractmethod
    def fit_transform_params(column: np.ndarray, backend: Any) -> Dict[str, Any]:
        pass


class ZScoreTransformer(NumberTransformer):
    def __init__(self, mean: float = None, std: float = None, **kwargs: dict):
        super().__init__()
        self.mu = float(mean) if mean is not None else mean
        self.sigma = float(std) if std is not None else std
        self.feature_name = kwargs.get(NAME, "")
        if self.sigma == 0:
            raise RuntimeError(
                f"Cannot apply zscore normalization to `{self.feature_name}` since it has a standard deviation of 0. "
                f"This is most likely because `{self.feature_name}` has a constant value of {self.mu} for all rows in "
                "the dataset. Consider removing this feature from your Ludwig config since it is not useful for "
                "your machine learning model."
            )

    def transform(self, x: np.ndarray) -> np.ndarray:
        return (x - self.mu) / self.sigma

    def inverse_transform(self, x: np.ndarray) -> np.ndarray:
        return x * self.sigma + self.mu

    def transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.mu) / self.sigma

    def inverse_transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.sigma + self.mu

    @staticmethod
    def fit_transform_params(column: np.ndarray, backend: "Backend") -> Dict[str, Any]:  # noqa
        compute = backend.df_engine.compute
        return {
            "mean": compute(column.astype(np.float32).mean()),
            "std": compute(column.astype(np.float32).std()),
        }


class MinMaxTransformer(NumberTransformer):
    def __init__(self, min: float = None, max: float = None, **kwargs: dict):
        super().__init__()
        self.min_value = float(min) if min is not None else min
        self.max_value = float(max) if max is not None else max
        if self.min_value is None or self.max_value is None:
            self.range = None
        else:
            self.range = self.max_value - self.min_value

    def transform(self, x: np.ndarray) -> np.ndarray:
        return (x - self.min_value) / self.range

    def inverse_transform(self, x: np.ndarray) -> np.ndarray:
        if self.range is None:
            raise ValueError("Numeric transformer needs to be instantiated with " "min and max values.")
        return x * self.range + self.min_value

    def transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.min_value) / self.range

    def inverse_transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        if self.range is None:
            raise ValueError("Numeric transformer needs to be instantiated with " "min and max values.")
        return x * self.range + self.min_value

    @staticmethod
    def fit_transform_params(column: np.ndarray, backend: "Backend") -> Dict[str, Any]:  # noqa
        compute = backend.df_engine.compute
        return {
            "min": compute(column.astype(np.float32).min()),
            "max": compute(column.astype(np.float32).max()),
        }


class InterQuartileTransformer(NumberTransformer):
    def __init__(self, q1: float = None, q2: float = None, q3: float = None, **kwargs: dict):
        super().__init__()
        self.q1 = float(q1) if q1 is not None else q1
        self.q2 = float(q2) if q2 is not None else q2
        self.q3 = float(q3) if q3 is not None else q3
        if self.q1 is None or self.q3 is None:
            self.interquartile_range = None
        else:
            self.interquartile_range = self.q3 - self.q1
        self.feature_name = kwargs.get(NAME, "")
        if self.interquartile_range == 0:
            raise RuntimeError(
                f"Cannot apply InterQuartileNormalization to `{self.feature_name}` since"
                "the interquartile range is 0, which will result in a ZeroDivisionError."
            )

    def transform(self, x: np.ndarray) -> np.ndarray:
        return (x - self.q2) / self.interquartile_range

    def inverse_transform(self, x: np.ndarray) -> np.ndarray:
        return x * self.interquartile_range + self.q2

    def transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.q2) / self.interquartile_range

    def inverse_transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.interquartile_range + self.q2

    @staticmethod
    def fit_transform_params(column: np.ndarray, backend: "Backend") -> Dict[str, Any]:  # noqa
        compute = backend.df_engine.compute
        return {
            "q1": compute(np.percentile(column.astype(np.float32), 25)),
            "q2": compute(np.percentile(column.astype(np.float32), 50)),
            "q3": compute(np.percentile(column.astype(np.float32), 75)),
        }


class Log1pTransformer(NumberTransformer):
    def __init__(self, **kwargs: dict):
        super().__init__()

    def transform(self, x: np.ndarray) -> np.ndarray:
        if np.any(x <= 0):
            raise ValueError(
                "One or more values are non-positive.  " "log1p normalization is defined only for positive values."
            )
        return np.log1p(x)

    def inverse_transform(self, x: np.ndarray) -> np.ndarray:
        return np.expm1(x)

    def transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return torch.log1p(x)

    def inverse_transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return torch.expm1(x)

    @staticmethod
    def fit_transform_params(column: np.ndarray, backend: "Backend") -> Dict[str, Any]:  # noqa
        return {}


class IdentityTransformer(NumberTransformer):
    def __init__(self, **kwargs):
        super().__init__()

    def transform(self, x: np.ndarray) -> np.ndarray:
        return x

    def inverse_transform(self, x: np.ndarray) -> np.ndarray:
        return x

    def transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return x

    def inverse_transform_inference(self, x: torch.Tensor) -> torch.Tensor:
        return x

    @staticmethod
    def fit_transform_params(column: np.ndarray, backend: "Backend") -> Dict[str, Any]:  # noqa
        return {}


numeric_transformation_registry = {
    "minmax": MinMaxTransformer,
    "zscore": ZScoreTransformer,
    "log1p": Log1pTransformer,
    "iq": InterQuartileTransformer,
    None: IdentityTransformer,
}


def get_transformer(metadata, preprocessing_parameters) -> NumberTransformer:
    return get_from_registry(
        preprocessing_parameters.get("normalization", None),
        numeric_transformation_registry,
    )(**metadata)


class _OutlierReplacer(torch.nn.Module):
    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.zscore_transformer = ZScoreTransformer(**metadata)
        self.outlier_threshold = metadata["preprocessing"].get("outlier_threshold")
        self.computed_outlier_fill_value = float(metadata["preprocessing"]["computed_outlier_fill_value"])

    def forward(self, v: torch.Tensor) -> torch.Tensor:
        outliers = self.zscore_transformer.transform_inference(v).abs().gt(self.outlier_threshold)
        v_masked = torch.masked_fill(v, outliers, torch.nan)

        v = torch.nan_to_num(v_masked, nan=self.computed_outlier_fill_value)
        return v.to(dtype=torch.float32)


class _NumberPreprocessing(torch.nn.Module):
    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.computed_fill_value = float(metadata["preprocessing"]["computed_fill_value"])
        self.numeric_transformer = get_transformer(metadata, metadata["preprocessing"])

        # Optional outlier replacement
        self.outlier_replacer = None
        if metadata["preprocessing"].get("outlier_strategy") is not None:
            self.outlier_replacer = _OutlierReplacer(metadata)

    def forward(self, v: TorchscriptPreprocessingInput) -> torch.Tensor:
        if not torch.jit.isinstance(v, torch.Tensor):
            raise ValueError(f"Unsupported input: {v}")

        v = torch.nan_to_num(v, nan=self.computed_fill_value)
        v = v.to(dtype=torch.float32)

        # Handle outliers if needed
        if self.outlier_replacer is not None:
            v = self.outlier_replacer(v)

        return self.numeric_transformer.transform_inference(v)


class _NumberPostprocessing(torch.nn.Module):
    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.numeric_transformer = get_transformer(metadata, metadata["preprocessing"])
        self.predictions_key = PREDICTIONS

    def forward(self, preds: Dict[str, torch.Tensor], feature_name: str) -> FeaturePostProcessingOutputDict:
        predictions = output_feature_utils.get_output_feature_tensor(preds, feature_name, self.predictions_key)

        return {self.predictions_key: self.numeric_transformer.inverse_transform_inference(predictions)}


class _NumberPredict(PredictModule):
    def __init__(self, clip):
        super().__init__()
        self.clip = clip

    def forward(self, inputs: Dict[str, torch.Tensor], feature_name: str) -> Dict[str, torch.Tensor]:
        logits = output_feature_utils.get_output_feature_tensor(inputs, feature_name, self.logits_key)
        predictions = logits

        if self.clip is not None:
            predictions = torch.clamp(logits, self.clip[0], self.clip[1])
            logger.debug(f"  clipped_predictions: {predictions}")

        return {self.predictions_key: predictions, self.logits_key: logits}


class NumberFeatureMixin(BaseFeatureMixin):
    @staticmethod
    def type():
        return NUMBER

    @staticmethod
    def cast_column(column, backend):
        return backend.df_engine.df_lib.to_numeric(column, errors="coerce").astype(np.float32)

    @staticmethod
    def get_feature_meta(
        column, preprocessing_parameters: PreprocessingConfigDict, backend, is_input_feature: bool
    ) -> FeatureMetadataDict:
        numeric_transformer: NumberTransformer = get_from_registry(
            preprocessing_parameters.get("normalization", None),
            numeric_transformation_registry,
        )

        params = numeric_transformer.fit_transform_params(column, backend)

        # Ensure mean and std are computed if we're removing outliers
        outlier_strategy = preprocessing_parameters.get("outlier_strategy")
        if outlier_strategy is not None and ("mean" not in params or "std" not in params):
            params.update(ZScoreTransformer.fit_transform_params(column, backend))

        return params

    @staticmethod
    def add_feature_data(
        feature_config,
        input_df,
        proc_df,
        metadata,
        preprocessing_parameters: PreprocessingConfigDict,
        backend,
        skip_save_processed_input,
    ):
        # Had to replace normalize() function due to issue #1911
        # this comment is to provide context for the change.
        # original code
        # def normalize(series: pd.Series) -> pd.Series:
        #     series = series.copy()
        #     numeric_transformer = get_transformer(metadata[feature_config[NAME]], preprocessing_parameters)
        #     series.update(numeric_transformer.transform(series.values))
        #     return series

        def normalize(series: pd.Series) -> pd.Series:
            _feature_metadata = copy.deepcopy(metadata[feature_config[NAME]])
            _feature_metadata.update({NAME: feature_config[NAME]})

            # retrieve request numeric transformer
            numeric_transformer = get_transformer(_feature_metadata, preprocessing_parameters)

            # transform input numeric values with specified transformer
            transformed_values = numeric_transformer.transform(series.values)

            # return transformed values with same index values as original series.
            return pd.Series(transformed_values, index=series.index)

        input_series = input_df[feature_config[COLUMN]].astype(np.float32)
        proc_df[feature_config[PROC_COLUMN]] = backend.df_engine.map_partitions(
            input_series, normalize, meta=input_series
        )

        return proc_df


class NumberInputFeature(NumberFeatureMixin, InputFeature):
    def __init__(self, input_feature_config: NumberInputFeatureConfig, encoder_obj=None, **kwargs):
        super().__init__(input_feature_config, **kwargs)
        input_feature_config.encoder.input_size = self.input_shape[-1]

        if encoder_obj:
            self.encoder_obj = encoder_obj
        else:
            self.encoder_obj = self.initialize_encoder(input_feature_config.encoder)

    def forward(self, inputs):
        assert isinstance(inputs, torch.Tensor)
        assert inputs.dtype == torch.float32 or inputs.dtype == torch.float64
        assert len(inputs.shape) == 1 or (len(inputs.shape) == 2 and inputs.shape[1] == 1)

        if len(inputs.shape) == 1:
            inputs = inputs[:, None]
        inputs_encoded = self.encoder_obj(inputs)

        return inputs_encoded

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([1])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self.encoder_obj.output_shape)

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        pass

    @staticmethod
    def get_schema_cls():
        return NumberInputFeatureConfig

    def create_sample_input(self, batch_size: int = 2):
        return torch.rand([batch_size])

    @classmethod
    def get_preproc_input_dtype(cls, metadata: TrainingSetMetadataDict) -> str:
        return "float32"

    @staticmethod
    def create_preproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _NumberPreprocessing(metadata)


class NumberOutputFeature(NumberFeatureMixin, OutputFeature):
    def __init__(
        self,
        output_feature_config: Union[NumberOutputFeatureConfig, Dict],
        output_features: Dict[str, OutputFeature],
        **kwargs,
    ):
        self.clip = output_feature_config.clip
        super().__init__(output_feature_config, output_features, **kwargs)
        self.decoder_obj = self.initialize_decoder(output_feature_config.decoder)
        self._setup_loss()
        self._setup_metrics()

    def logits(self, inputs, **kwargs):  # hidden
        hidden = inputs[HIDDEN]
        return self.decoder_obj(hidden)

    def create_predict_module(self) -> PredictModule:
        if getattr(self, "clip", None) and not (isinstance(self.clip, (list, tuple)) and len(self.clip) == 2):
            raise ValueError(
                f"The clip parameter of {self.feature_name} is {self.clip}. "
                f"It must be a list or a tuple of length 2."
            )
        return _NumberPredict(getattr(self, "clip", None))

    def get_prediction_set(self):
        return {PREDICTIONS, LOGITS}

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.decoder_obj.config.input_size])

    @classmethod
    def get_output_dtype(cls):
        return torch.float32

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([1])

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        pass

    @staticmethod
    def calculate_overall_stats(predictions, targets, metadata):
        # no overall stats, just return empty dictionary
        return {}

    def postprocess_predictions(
        self,
        predictions,
        metadata,
    ):
        predictions_col = f"{self.feature_name}_{PREDICTIONS}"
        if predictions_col in predictions:
            # as needed convert predictions make to original value space
            numeric_transformer = get_from_registry(
                metadata["preprocessing"].get("normalization", None),
                numeric_transformation_registry,
            )(**metadata)
            predictions[predictions_col] = predictions[predictions_col].map(
                lambda pred: numeric_transformer.inverse_transform(pred)
            )

        return predictions

    @staticmethod
    def get_schema_cls():
        return NumberOutputFeatureConfig

    @classmethod
    def get_postproc_output_dtype(cls, metadata: TrainingSetMetadataDict) -> str:
        return "float32"

    @staticmethod
    def create_postproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _NumberPostprocessing(metadata)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
import warnings
from collections import Counter
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torchvision.transforms import functional as F
from torchvision.transforms.functional import normalize

from ludwig.constants import (
    CHECKSUM,
    COLUMN,
    ENCODER,
    HEIGHT,
    IMAGE,
    IMAGENET1K,
    INFER_IMAGE_DIMENSIONS,
    INFER_IMAGE_MAX_HEIGHT,
    INFER_IMAGE_MAX_WIDTH,
    INFER_IMAGE_SAMPLE_SIZE,
    NAME,
    NUM_CHANNELS,
    PREPROCESSING,
    PROC_COLUMN,
    REQUIRES_EQUAL_DIMENSIONS,
    SRC,
    TRAINING,
    TYPE,
    WIDTH,
)
from ludwig.data.cache.types import wrap
from ludwig.encoders.image.torchvision import TVModelVariant
from ludwig.features.base_feature import BaseFeatureMixin, InputFeature
from ludwig.schema.features.augmentation.base import BaseAugmentationConfig
from ludwig.schema.features.augmentation.image import (
    RandomBlurConfig,
    RandomBrightnessConfig,
    RandomContrastConfig,
    RandomHorizontalFlipConfig,
    RandomRotateConfig,
    RandomVerticalFlipConfig,
)
from ludwig.schema.features.image_feature import ImageInputFeatureConfig
from ludwig.types import FeatureMetadataDict, PreprocessingConfigDict, TrainingSetMetadataDict
from ludwig.utils.augmentation_utils import get_augmentation_op, register_augmentation_op
from ludwig.utils.data_utils import get_abs_path
from ludwig.utils.dataframe_utils import is_dask_series_or_df
from ludwig.utils.fs_utils import has_remote_protocol, upload_h5
from ludwig.utils.image_utils import (
    get_gray_default_image,
    grayscale,
    is_torchvision_encoder,
    num_channels_in_image,
    read_image_from_bytes_obj,
    read_image_from_path,
    resize_image,
    ResizeChannels,
    torchvision_model_registry,
)
from ludwig.utils.misc_utils import set_default_value
from ludwig.utils.types import Series, TorchscriptPreprocessingInput

# constants used for Ludwig image preprocessing
IMAGENET1K_MEAN = [0.485, 0.456, 0.406]
IMAGENET1K_STD = [0.229, 0.224, 0.225]


logger = logging.getLogger(__name__)


###
# Image specific augmentation operations
###
@register_augmentation_op(name="random_vertical_flip", features=IMAGE)
class RandomVFlip(torch.nn.Module):
    def __init__(
        self,
        config: RandomVerticalFlipConfig,
    ):
        super().__init__()

    def forward(self, imgs):
        if torch.rand(1) < 0.5:
            imgs = F.vflip(imgs)

        return imgs


@register_augmentation_op(name="random_horizontal_flip", features=IMAGE)
class RandomHFlip(torch.nn.Module):
    def __init__(
        self,
        config: RandomHorizontalFlipConfig,
    ):
        super().__init__()

    def forward(self, imgs):
        if torch.rand(1) < 0.5:
            imgs = F.hflip(imgs)

        return imgs


@register_augmentation_op(name="random_rotate", features=IMAGE)
class RandomRotate(torch.nn.Module):
    def __init__(self, config: RandomRotateConfig):
        super().__init__()
        self.degree = config.degree

    def forward(self, imgs):
        if torch.rand(1) < 0.5:
            # map angle to interval (-degree, +degree)
            angle = (torch.rand(1) * 2 * self.degree - self.degree).item()
            return F.rotate(imgs, angle)
        else:
            return imgs


@register_augmentation_op(name="random_contrast", features=IMAGE)
class RandomContrast(torch.nn.Module):
    def __init__(self, config: RandomContrastConfig):
        super().__init__()
        self.min_contrast = config.min
        self.contrast_adjustment_range = config.max - config.min

    def forward(self, imgs):
        if torch.rand(1) < 0.5:
            # random contrast adjustment
            adjust_factor = (torch.rand(1) * self.contrast_adjustment_range + self.min_contrast).item()
            return F.adjust_contrast(imgs, adjust_factor)
        else:
            return imgs


@register_augmentation_op(name="random_brightness", features=IMAGE)
class RandomBrightness(torch.nn.Module):
    def __init__(self, config: RandomBrightnessConfig):
        super().__init__()
        self.min_brightness = config.min
        self.brightness_adjustment_range = config.max - config.min

    def forward(self, imgs):
        if torch.rand(1) < 0.5:
            # random contrast adjustment
            adjust_factor = (torch.rand(1) * self.brightness_adjustment_range + self.min_brightness).item()
            return F.adjust_brightness(imgs, adjust_factor)
        else:
            return imgs


@register_augmentation_op(name="random_blur", features=IMAGE)
class RandomBlur(torch.nn.Module):
    def __init__(self, config: RandomBlurConfig):
        super().__init__()
        self.kernel_size = [config.kernel_size, config.kernel_size]

    def forward(self, imgs):
        if torch.rand(1) < 0.5:
            imgs = F.gaussian_blur(imgs, self.kernel_size)

        return imgs


class ImageAugmentation(torch.nn.Module):
    def __init__(
        self,
        augmentation_list: List[BaseAugmentationConfig],
        normalize_mean: Optional[List[float]] = None,
        normalize_std: Optional[List[float]] = None,
    ):
        super().__init__()

        # TODO: change to debug level before merging
        logger.info(f"Creating Augmentation pipline: {augmentation_list}")

        self.normalize_mean = normalize_mean
        self.normalize_std = normalize_std

        if self.training:
            self.augmentation_steps = torch.nn.Sequential()
            for aug_config in augmentation_list:
                try:
                    aug_op = get_augmentation_op(IMAGE, aug_config.type)
                    self.augmentation_steps.append(aug_op(aug_config))
                except KeyError:
                    raise ValueError(f"Invalid augmentation operation specification: {aug_config}")
        else:
            # TODO: should this raise an exception if not in training mode?
            self.augmentation_steps = None

    def forward(self, imgs):
        if self.augmentation_steps:
            # convert from float to uint8 values - this is required for the augmentation
            imgs = self._convert_back_to_uint8(imgs)

            logger.debug(f"Executing augmentation pipeline steps:\n{self.augmentation_steps}")
            imgs = self.augmentation_steps(imgs)

            # convert back to float32 values and renormalize if needed
            imgs = self._renormalize_image(imgs)

        return imgs

    # function to partially undo the TorchVision ImageClassification transformation.
    #  back out the normalization step and convert from float32 to uint8 dtype
    #  to make the tensor displayable as an image
    #  crop size remains the same
    def _convert_back_to_uint8(self, images):
        if self.normalize_mean:
            mean = torch.as_tensor(self.normalize_mean, dtype=torch.float32).view(-1, 1, 1)
            std = torch.as_tensor(self.normalize_std, dtype=torch.float32).view(-1, 1, 1)
            return images.mul(std).add(mean).mul(255.0).type(torch.uint8)
        else:
            return images.mul(255.0).type(torch.uint8)

    # function to redo part of the TorchVision ImageClassification transformation.
    #  convert uint8 to float32
    #  apply the imagenet1k normalization
    def _renormalize_image(self, images):
        if self.normalize_mean:
            mean = torch.as_tensor(self.normalize_mean, dtype=torch.float32).view(-1, 1, 1)
            std = torch.as_tensor(self.normalize_std, dtype=torch.float32).view(-1, 1, 1)
            return images.type(torch.float32).div(255.0).sub(mean).div(std)
        else:
            return images.type(torch.float32).div(255.0)


@dataclass
class ImageTransformMetadata:
    height: int
    width: int
    num_channels: int


def _get_torchvision_transform(
    torchvision_parameters: TVModelVariant,
) -> Tuple[torch.nn.Module, ImageTransformMetadata]:
    """Returns a torchvision transform that is compatible with the model variant.

    Note that the raw torchvision transform is not returned. Instead, a Sequential module that includes
    image resizing is returned. This is because the raw torchvision transform assumes that the input image has
    three channels, which is not always the case with images input into Ludwig.

    Args:
        torchvision_parameters: The parameters for the torchvision model variant.
    Returns:
        (torchvision_transform, transform_metadata): A torchvision transform and the metadata for the transform.
    """
    torchvision_transform_raw = torchvision_parameters.model_weights.DEFAULT.transforms()
    torchvision_transform = torch.nn.Sequential(
        ResizeChannels(num_channels=3),
        torchvision_transform_raw,
    )
    transform_metadata = ImageTransformMetadata(
        height=torchvision_transform_raw.crop_size[0],
        width=torchvision_transform_raw.crop_size[0],
        num_channels=len(torchvision_transform_raw.mean),
    )
    return (torchvision_transform, transform_metadata)


def _get_torchvision_parameters(model_type: str, model_variant: str) -> TVModelVariant:
    return torchvision_model_registry.get(model_type).get(model_variant)


class _ImagePreprocessing(torch.nn.Module):
    """Torchscript-enabled version of preprocessing done by ImageFeatureMixin.add_feature_data."""

    def __init__(
        self,
        metadata: TrainingSetMetadataDict,
        torchvision_transform: Optional[torch.nn.Module] = None,
        transform_metadata: Optional[ImageTransformMetadata] = None,
    ):
        super().__init__()

        self.resize_method = metadata["preprocessing"]["resize_method"]
        self.torchvision_transform = torchvision_transform
        if transform_metadata is not None:
            self.height = transform_metadata.height
            self.width = transform_metadata.width
            self.num_channels = transform_metadata.num_channels
        else:
            self.height = metadata["preprocessing"]["height"]
            self.width = metadata["preprocessing"]["width"]
            self.num_channels = metadata["preprocessing"]["num_channels"]

    def forward(self, v: TorchscriptPreprocessingInput) -> torch.Tensor:
        """Takes a list of images and adjusts the size and number of channels as specified in the metadata.

        If `v` is already a torch.Tensor, we assume that the images are already preprocessed to be the same size.
        """
        # Nested conditional is a workaround to short-circuit boolean evaluation.
        if not torch.jit.isinstance(v, List[torch.Tensor]):
            if not torch.jit.isinstance(v, torch.Tensor):
                raise ValueError(f"Unsupported input: {v}")

        if self.torchvision_transform is not None:
            # perform pre-processing for torchvision pretrained model encoders
            if torch.jit.isinstance(v, List[torch.Tensor]):
                imgs = [self.torchvision_transform(img) for img in v]
            else:
                # convert batch of image tensors to a list and then run torchvision pretrained
                # model transforms on each image
                imgs = [self.torchvision_transform(img) for img in torch.unbind(v)]

            # collect the list of images into a batch
            imgs_stacked = torch.stack(imgs)
        else:
            # perform pre-processing for Ludwig defined image encoders
            if torch.jit.isinstance(v, List[torch.Tensor]):
                imgs = [resize_image(img, (self.height, self.width), self.resize_method) for img in v]
                imgs_stacked = torch.stack(imgs)
            else:
                imgs_stacked = v

            _, num_channels, height, width = imgs_stacked.shape

            # Ensure images are the size expected by the model
            if height != self.height or width != self.width:
                imgs_stacked = resize_image(imgs_stacked, (self.height, self.width), self.resize_method)

            # Ensures images have the number of channels expected by the model
            if num_channels != self.num_channels:
                if self.num_channels == 1:
                    imgs_stacked = grayscale(imgs_stacked)
                elif num_channels < self.num_channels:
                    extra_channels = self.num_channels - num_channels
                    imgs_stacked = torch.nn.functional.pad(imgs_stacked, [0, 0, 0, 0, 0, extra_channels])
                else:
                    raise ValueError(
                        f"Number of channels cannot be reconciled. metadata.num_channels = "
                        f"{self.num_channels}, but imgs.shape[1] = {num_channels}"
                    )

            imgs_stacked = imgs_stacked.type(torch.float32) / 255

        return imgs_stacked


class ImageFeatureMixin(BaseFeatureMixin):
    @staticmethod
    def type():
        return IMAGE

    @staticmethod
    def cast_column(column, backend):
        return column

    @staticmethod
    def get_feature_meta(
        column, preprocessing_parameters: PreprocessingConfigDict, backend, is_input_feature: bool
    ) -> FeatureMetadataDict:
        return {PREPROCESSING: preprocessing_parameters}

    @staticmethod
    def _read_image_if_bytes_obj_and_resize(
        img_entry: Union[bytes, torch.Tensor, np.ndarray],
        img_width: int,
        img_height: int,
        should_resize: bool,
        num_channels: int,
        resize_method: str,
        user_specified_num_channels: bool,
        standardize_image: str,
    ) -> Optional[np.ndarray]:
        """
        :param img_entry Union[bytes, torch.Tensor, np.ndarray]: if str file path to the
            image else torch.Tensor of the image itself
        :param img_width: expected width of the image
        :param img_height: expected height of the image
        :param should_resize: Should the image be resized?
        :param resize_method: type of resizing method
        :param num_channels: expected number of channels in the first image
        :param user_specified_num_channels: did the user specify num channels?
        :param standardize_image: specifies whether to standarize image with imagenet1k specifications
        :return: image object as a numpy array

        Helper method to read and resize an image according to model definition.
        If the user doesn't specify a number of channels, we use the first image
        in the dataset as the source of truth. If any image in the dataset
        doesn't have the same number of channels as the first image,
        raise an exception.

        If the user specifies a number of channels, we try to convert all the
        images to the specifications by dropping channels/padding 0 channels
        """

        if isinstance(img_entry, bytes):
            img = read_image_from_bytes_obj(img_entry, num_channels)
        elif isinstance(img_entry, str):
            img = read_image_from_path(img_entry, num_channels)
        elif isinstance(img_entry, np.ndarray):
            img = torch.from_numpy(np.array(img_entry, copy=True)).permute(2, 0, 1)
        else:
            img = img_entry

        if not isinstance(img, torch.Tensor):
            warnings.warn(f"Image with value {img} cannot be read")
            return None

        img_num_channels = num_channels_in_image(img)
        # Convert to grayscale if needed.
        if num_channels == 1 and img_num_channels != 1:
            img = grayscale(img)
            img_num_channels = 1

        if should_resize:
            img = resize_image(img, (img_height, img_width), resize_method)

        if user_specified_num_channels:
            # Number of channels is specified by the user
            # img_padded = np.zeros((img_height, img_width, num_channels),
            #                       dtype=np.uint8)
            # min_num_channels = min(num_channels, img_num_channels)
            # img_padded[:, :, :min_num_channels] = img[:, :, :min_num_channels]
            # img = img_padded
            if num_channels > img_num_channels:
                extra_channels = num_channels - img_num_channels
                img = torch.nn.functional.pad(img, [0, 0, 0, 0, 0, extra_channels])

            if img_num_channels != num_channels:
                logger.warning(
                    "Image has {} channels, where as {} "
                    "channels are expected. Dropping/adding channels "
                    "with 0s as appropriate".format(img_num_channels, num_channels)
                )
        else:
            # If the image isn't like the first image, raise exception
            if img_num_channels != num_channels:
                raise ValueError(
                    "Image has {} channels, unlike the first image, which "
                    "has {} channels. Make sure all the images have the same "
                    "number of channels or use the num_channels property in "
                    "image preprocessing".format(img_num_channels, num_channels)
                )

        if img.shape[1] != img_height or img.shape[2] != img_width:
            raise ValueError(
                "Images are not of the same size. "
                "Expected size is {}, "
                "current image size is {}."
                "Images are expected to be all of the same size "
                "or explicit image width and height are expected "
                "to be provided. "
                "Additional information: "
                "https://ludwig-ai.github.io/ludwig-docs/latest/configuration/features/image_features"
                "#image-features-preprocessing".format([img_height, img_width, num_channels], img.shape)
            )

        # casting and rescaling
        img = img.type(torch.float32) / 255

        if standardize_image == IMAGENET1K:
            img = normalize(img, mean=IMAGENET1K_MEAN, std=IMAGENET1K_STD)

        return img.numpy()

    @staticmethod
    def _read_image_with_pretrained_transform(
        img_entry: Union[bytes, torch.Tensor, np.ndarray],
        transform_fn: Callable,
    ) -> Optional[np.ndarray]:

        if isinstance(img_entry, bytes):
            img = read_image_from_bytes_obj(img_entry)
        elif isinstance(img_entry, str):
            img = read_image_from_path(img_entry)
        elif isinstance(img_entry, np.ndarray):
            img = torch.from_numpy(img_entry).permute(2, 0, 1)
        else:
            img = img_entry

        if not isinstance(img, torch.Tensor):
            warnings.warn(f"Image with value {img} cannot be read")
            return None

        img = transform_fn(img)

        return img.numpy()

    @staticmethod
    def _set_image_and_height_equal_for_encoder(
        width: int, height: int, preprocessing_parameters: dict, encoder_type: str
    ) -> Tuple[int, int]:
        """Some pretrained image encoders require images with the same dimension, or images with a specific width
        and heigh values. The returned width and height are set based on compatibility with the downstream encoder
        using the encoder parameters for the feature.

        Args:
            width: Represents the width of the image. This is either specified in the user config, or inferred using
                a sample of images.
            height: Represents the height of the image. This is either specified in the user config, or inferred using
                a sample of images.
            preprocessing_parameters: Parameters defining how the image feature should be preprocessed
            encoder_type: The name of the encoder

        Return:
            (width, height) Updated width and height so that they are equal
        """

        if preprocessing_parameters[REQUIRES_EQUAL_DIMENSIONS] and height != width:
            width = height = min(width, height)
            # Update preprocessing parameters dictionary to reflect new height and width values
            preprocessing_parameters["width"] = width
            preprocessing_parameters["height"] = height
            logger.info(
                f"Set image feature height and width to {width} to be compatible with" f" {encoder_type} encoder."
            )
        return width, height

    @staticmethod
    def _infer_image_size(
        image_sample: List[torch.Tensor],
        max_height: int,
        max_width: int,
        preprocessing_parameters: dict,
        encoder_type: str,
    ) -> Tuple[int, int]:
        """Infers the size to use from a group of images. The returned height will be the average height of images
        in image_sample rounded to the nearest integer, or max_height. Likewise for width.

        Args:
            image_sample: Sample of images to use to infer image size. Must be formatted as [channels, height, width].
            max_height: Maximum height.
            max_width: Maximum width.
            preprocessing_parameters: Parameters defining how the image feature should be preprocessed
            encoder_type: The name of the encoder

        Return:
            (height, width) The inferred height and width.
        """

        height_avg = sum(x.shape[1] for x in image_sample) / len(image_sample)
        width_avg = sum(x.shape[2] for x in image_sample) / len(image_sample)
        height = min(int(round(height_avg)), max_height)
        width = min(int(round(width_avg)), max_width)

        # Update height and width if the downstream encoder requires images
        # with  the same dimension or specific width and height values
        width, height = ImageFeatureMixin._set_image_and_height_equal_for_encoder(
            width, height, preprocessing_parameters, encoder_type
        )

        logger.debug(f"Inferring height: {height} and width: {width}")
        return height, width

    @staticmethod
    def _infer_number_of_channels(image_sample: List[torch.Tensor]):
        """Infers the channel depth to use from a group of images.

        We make the assumption that the majority of datasets scraped from the web will be RGB, so if we get a mixed bag
        of images we should default to that. However, if the majority of the sample images have a specific channel depth
        (other than 3) this is probably intentional so we keep it, but log an info message.
        """
        n_images = len(image_sample)
        channel_frequency = Counter([num_channels_in_image(x) for x in image_sample])
        if channel_frequency[1] > n_images / 2:
            # If the majority of images in sample are 1 channel, use 1.
            num_channels = 1
        elif channel_frequency[2] > n_images / 2:
            # If the majority of images in sample are 2 channel, use 2.
            num_channels = 2
        elif channel_frequency[4] > n_images / 2:
            # If the majority of images in sample are 4 channel, use 4.
            num_channels = 4
        else:
            # Default case: use 3 channels.
            num_channels = 3
        logger.info(f"Inferring num_channels from the first {n_images} images.")
        logger.info("\n".join([f"  images with {k} channels: {v}" for k, v in sorted(channel_frequency.items())]))
        if num_channels == max(channel_frequency, key=channel_frequency.get):
            logger.info(
                f"Using {num_channels} channels because it is the majority in sample. If an image with"
                f" a different depth is read, will attempt to convert to {num_channels} channels."
            )
        else:
            logger.info(f"Defaulting to {num_channels} channels.")
        logger.info(
            "To explicitly set the number of channels, define num_channels in the preprocessing dictionary of "
            "the image input feature config."
        )
        return num_channels

    @staticmethod
    def _finalize_preprocessing_parameters(
        preprocessing_parameters: dict,
        encoder_type: str,
        column: Series,
    ) -> Tuple:
        """Helper method to determine the height, width and number of channels for preprocessing the image data.

        This is achieved by looking at the parameters provided by the user. When there are some missing parameters, we
        fall back on to the first image in the dataset. The assumption being that all the images in the data are
        expected be of the same size with the same number of channels.

        Args:
            preprocessing_parameters: Parameters defining how the image feature should be preprocessed
            encoder_type: The name of the encoder
            column: The data itself. Can be a Pandas, Modin or Dask series.
        """

        explicit_height_width = preprocessing_parameters[HEIGHT] or preprocessing_parameters[WIDTH]
        explicit_num_channels = NUM_CHANNELS in preprocessing_parameters and preprocessing_parameters[NUM_CHANNELS]

        if preprocessing_parameters[INFER_IMAGE_DIMENSIONS] and not (explicit_height_width and explicit_num_channels):
            sample_size = min(len(column), preprocessing_parameters[INFER_IMAGE_SAMPLE_SIZE])
        else:
            sample_size = 1  # Take first image

        sample = []
        sample_num_bytes = []
        failed_entries = []
        for image_entry in column.head(sample_size):
            if isinstance(image_entry, str):
                # Tries to read image as PNG or numpy file from the path.
                image, num_bytes = read_image_from_path(image_entry, return_num_bytes=True)
                if num_bytes is not None:
                    sample_num_bytes.append(num_bytes)
            else:
                image = image_entry

            if isinstance(image, torch.Tensor):
                sample.append(image)
            elif isinstance(image, np.ndarray):
                sample.append(torch.from_numpy(image).permute(2, 0, 1))
            else:
                failed_entries.append(image_entry)
        if len(sample) == 0:
            failed_entries_repr = "\n\t- ".join(failed_entries)
            raise ValueError(
                f"Images dimensions cannot be inferred. Failed to read {sample_size} images as samples:\n\t- "
                f"{failed_entries_repr}."
            )

        should_resize = False
        if explicit_height_width:
            should_resize = True
            try:
                height = int(preprocessing_parameters[HEIGHT])
                width = int(preprocessing_parameters[WIDTH])
                # Update height and width if the downstream encoder requires images
                # with the same dimension or specific width and height values
                width, height = ImageFeatureMixin._set_image_and_height_equal_for_encoder(
                    width, height, preprocessing_parameters, encoder_type
                )
            except ValueError as e:
                raise ValueError("Image height and width must be set and have " "positive integer values: " + str(e))
            if height <= 0 or width <= 0:
                raise ValueError("Image height and width must be positive integers")
        else:
            # User hasn't specified height and width.
            # Default to inferring from sample or first image.
            if preprocessing_parameters[INFER_IMAGE_DIMENSIONS]:
                should_resize = True
                height, width = ImageFeatureMixin._infer_image_size(
                    sample,
                    max_height=preprocessing_parameters[INFER_IMAGE_MAX_HEIGHT],
                    max_width=preprocessing_parameters[INFER_IMAGE_MAX_WIDTH],
                    preprocessing_parameters=preprocessing_parameters,
                    encoder_type=encoder_type,
                )
            else:
                raise ValueError(
                    "Explicit image width/height are not set, infer_image_dimensions is false, "
                    "and first image cannot be read, so image dimensions are unknown"
                )

        if explicit_num_channels:
            # User specified num_channels in the model/feature config
            user_specified_num_channels = True
            num_channels = preprocessing_parameters[NUM_CHANNELS]
        else:
            user_specified_num_channels = False
            if preprocessing_parameters[INFER_IMAGE_DIMENSIONS]:
                user_specified_num_channels = True
                num_channels = ImageFeatureMixin._infer_number_of_channels(sample)
            elif len(sample) > 0:
                num_channels = num_channels_in_image(sample[0])
            else:
                raise ValueError(
                    "Explicit image num channels is not set, infer_image_dimensions is false, "
                    "and first image cannot be read, so image num channels is unknown"
                )

        assert isinstance(num_channels, int), ValueError("Number of image channels needs to be an integer")

        average_file_size = np.mean(sample_num_bytes) if sample_num_bytes else None

        standardize_image = preprocessing_parameters["standardize_image"]
        if standardize_image == "imagenet1k" and num_channels != 3:
            warnings.warn(
                f"'standardize_image=imagenet1k' is defined only for 'num_channels=3' but "
                f"detected 'num_channels={num_channels}'.  For this situation setting 'standardize_image=None'.",
                RuntimeWarning,
            )
            standardize_image = None

        return (
            should_resize,
            width,
            height,
            num_channels,
            user_specified_num_channels,
            average_file_size,
            standardize_image,
        )

    @staticmethod
    def add_feature_data(
        feature_config,
        input_df,
        proc_df,
        metadata,
        preprocessing_parameters: PreprocessingConfigDict,
        backend,
        skip_save_processed_input,
    ):
        set_default_value(feature_config[PREPROCESSING], "in_memory", preprocessing_parameters["in_memory"])

        name = feature_config[NAME]
        column = input_df[feature_config[COLUMN]]
        encoder_type = feature_config[ENCODER][TYPE]

        src_path = None
        if SRC in metadata:
            src_path = os.path.dirname(os.path.abspath(metadata.get(SRC)))
        abs_path_column = backend.df_engine.map_objects(
            column,
            lambda row: get_abs_path(src_path, row) if isinstance(row, str) and not has_remote_protocol(row) else row,
        )

        # determine if specified encoder is a torchvision model
        model_type = feature_config[ENCODER].get("type", None)
        model_variant = feature_config[ENCODER].get("model_variant")
        if model_variant:
            torchvision_parameters = _get_torchvision_parameters(model_type, model_variant)
        else:
            torchvision_parameters = None

        if torchvision_parameters:
            logger.warning(
                f"Using the transforms specified for the torchvision model {model_type} {model_variant} "
                f"This includes setting the number of channels is 3 and resizing the image to the needs of the model."
            )

            torchvision_transform, transform_metadata = _get_torchvision_transform(torchvision_parameters)

            # torchvision_parameters is not None
            # perform torchvision model transformations
            read_image_if_bytes_obj_and_resize = partial(
                ImageFeatureMixin._read_image_with_pretrained_transform,
                transform_fn=torchvision_transform,
            )
            average_file_size = None

            # save weight specification in preprocessing section
            preprocessing_parameters[
                "torchvision_model_default_weights"
            ] = f"{torchvision_parameters.model_weights.DEFAULT}"

            # add torchvision model id to preprocessing section for torchscript
            preprocessing_parameters["torchvision_model_type"] = model_type
            preprocessing_parameters["torchvision_model_variant"] = model_variant

            # get required setup parameters for in_memory = False processing
            height = transform_metadata.height
            width = transform_metadata.width
            num_channels = transform_metadata.num_channels
        else:
            # torchvision_parameters is None
            # perform Ludwig specified transformations
            (
                should_resize,
                width,
                height,
                num_channels,
                user_specified_num_channels,
                average_file_size,
                standardize_image,
            ) = ImageFeatureMixin._finalize_preprocessing_parameters(
                preprocessing_parameters, encoder_type, abs_path_column
            )

            metadata[name][PREPROCESSING]["height"] = height
            metadata[name][PREPROCESSING]["width"] = width
            metadata[name][PREPROCESSING]["num_channels"] = num_channels

            read_image_if_bytes_obj_and_resize = partial(
                ImageFeatureMixin._read_image_if_bytes_obj_and_resize,
                img_width=width,
                img_height=height,
                should_resize=should_resize,
                num_channels=num_channels,
                resize_method=preprocessing_parameters["resize_method"],
                user_specified_num_channels=user_specified_num_channels,
                standardize_image=standardize_image,
            )

        # TODO: alternatively use get_average_image() for unreachable images
        default_image = get_gray_default_image(num_channels, height, width)
        metadata[name]["reshape"] = (num_channels, height, width)

        in_memory = feature_config[PREPROCESSING]["in_memory"]
        if in_memory or skip_save_processed_input:

            proc_col = backend.read_binary_files(
                abs_path_column, map_fn=read_image_if_bytes_obj_and_resize, file_size=average_file_size
            )

            num_failed_image_reads = (
                proc_col.isna().sum().compute() if is_dask_series_or_df(proc_col, backend) else proc_col.isna().sum()
            )

            proc_col = backend.df_engine.map_objects(
                proc_col, lambda row: default_image if not isinstance(row, np.ndarray) else row
            )

            proc_df[feature_config[PROC_COLUMN]] = proc_col
        else:
            num_images = len(abs_path_column)
            num_failed_image_reads = 0

            data_fp = backend.cache.get_cache_path(wrap(metadata.get(SRC)), metadata.get(CHECKSUM), TRAINING)
            with upload_h5(data_fp) as h5_file:
                # todo future add multiprocessing/multithreading
                image_dataset = h5_file.create_dataset(
                    feature_config[PROC_COLUMN] + "_data", (num_images, num_channels, height, width), dtype=np.float32
                )
                for i, img_entry in enumerate(abs_path_column):
                    res = read_image_if_bytes_obj_and_resize(img_entry)
                    if isinstance(res, np.ndarray):
                        image_dataset[i, :height, :width, :] = res
                    else:
                        logger.warning(f"Failed to read image {img_entry} while preprocessing feature `{name}`. ")
                        image_dataset[i, :height, :width, :] = default_image
                        num_failed_image_reads += 1
                h5_file.flush()

            proc_df[feature_config[PROC_COLUMN]] = np.arange(num_images)

        if num_failed_image_reads > 0:
            logger.warning(
                f"Failed to read {num_failed_image_reads} images while preprocessing feature `{name}`. "
                "Using default image for these rows in the dataset."
            )

        return proc_df


class ImageInputFeature(ImageFeatureMixin, InputFeature):
    def __init__(self, input_feature_config: ImageInputFeatureConfig, encoder_obj=None, **kwargs):
        super().__init__(input_feature_config, **kwargs)

        if encoder_obj:
            self.encoder_obj = encoder_obj
        else:
            self.encoder_obj = self.initialize_encoder(input_feature_config.encoder)

        # set up for augmentation if it is enabled
        if input_feature_config.augmentation:
            # assume no image normalize is required
            normalize_mean = normalize_std = None

            # determine if specified encoder is a torchvision model
            if is_torchvision_encoder(self.encoder_obj):
                # encoder is a torchvision model
                normalize_mean = self.encoder_obj.normalize_mean
                normalize_std = self.encoder_obj.normalize_std
            else:
                # encoder is a Ludwig encoder, determine if standardize_image is set to IMAGENET1K
                if input_feature_config.preprocessing.standardize_image == IMAGENET1K:
                    normalize_mean = IMAGENET1K_MEAN
                    normalize_std = IMAGENET1K_STD

            # create augmentation pipeline object
            self.augmentation_pipeline = ImageAugmentation(
                input_feature_config.augmentation,
                normalize_mean,
                normalize_std,
            )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        assert isinstance(inputs, torch.Tensor), f"inputs to image feature must be a torch tensor, got {type(inputs)}"
        assert inputs.dtype in [torch.float32], f"inputs to image feature must be a float32 tensor, got {inputs.dtype}"

        inputs_encoded = self.encoder_obj(inputs)

        return inputs_encoded

    @property
    def input_dtype(self):
        return torch.float32

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self.encoder_obj.input_shape)

    @property
    def output_shape(self) -> torch.Size:
        return self.encoder_obj.output_shape

    def update_config_after_module_init(self, feature_config):
        if is_torchvision_encoder(self.encoder_obj):
            # update feature preprocessing parameters to reflect used in torchvision pretrained model
            # Note: image height and width is determined by the encoder crop_size attribute.  Source of this
            # attribute is from the torchvision.transforms._presets.ImageClassification class.  This class stores
            # crop_size as a single element list.  the single element in this list is used to set both the height
            # and width of an image.
            feature_config.preprocessing.height = self.encoder_obj.crop_size[0]
            feature_config.preprocessing.width = self.encoder_obj.crop_size[0]
            feature_config.preprocessing.num_channels = self.encoder_obj.num_channels

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        for key in ["height", "width", "num_channels", "standardize_image"]:
            if hasattr(feature_config.encoder, key):
                setattr(feature_config.encoder, key, feature_metadata[PREPROCESSING][key])

    @staticmethod
    def get_schema_cls():
        return ImageInputFeatureConfig

    @staticmethod
    def create_preproc_module(metadata: Dict[str, Any]) -> torch.nn.Module:
        model_type = metadata["preprocessing"].get("torchvision_model_type")
        model_variant = metadata["preprocessing"].get("torchvision_model_variant")
        if model_variant:
            torchvision_parameters = _get_torchvision_parameters(model_type, model_variant)
        else:
            torchvision_parameters = None

        if torchvision_parameters:
            torchvision_transform, transform_metadata = _get_torchvision_transform(torchvision_parameters)
        else:
            torchvision_transform = None
            transform_metadata = None

        return _ImagePreprocessing(
            metadata, torchvision_transform=torchvision_transform, transform_metadata=transform_metadata
        )

    def get_augmentation_pipeline(self):
        return self.augmentation_pipeline



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import logging
import sys
from typing import List, Optional, Union

import pandas as pd

from ludwig.api import LudwigModel
from ludwig.backend import ALL_BACKENDS, Backend, initialize_backend
from ludwig.callbacks import Callback
from ludwig.constants import FULL, TEST, TRAINING, VALIDATION
from ludwig.contrib import add_contrib_callback_args
from ludwig.globals import LUDWIG_VERSION
from ludwig.utils.print_utils import get_logging_level_registry, print_ludwig

logger = logging.getLogger(__name__)


def evaluate_cli(
    model_path: str,
    dataset: Union[str, dict, pd.DataFrame] = None,
    data_format: str = None,
    split: str = FULL,
    batch_size: int = 128,
    skip_save_unprocessed_output: bool = False,
    skip_save_predictions: bool = False,
    skip_save_eval_stats: bool = False,
    skip_collect_predictions: bool = False,
    skip_collect_overall_stats: bool = False,
    output_directory: str = "results",
    gpus: Union[str, int, List[int]] = None,
    gpu_memory_limit: Optional[float] = None,
    allow_parallel_threads: bool = True,
    callbacks: List[Callback] = None,
    backend: Union[Backend, str] = None,
    logging_level: int = logging.INFO,
    **kwargs,
) -> None:
    """Loads pre-trained model and evaluates its performance by comparing the predictions against ground truth.

     # Inputs

     :param model_path: (str) filepath to pre-trained model.
     :param dataset: (Union[str, dict, pandas.DataFrame], default: `None`)
         source containing the entire dataset to be used in the evaluation.
     :param data_format: (str, default: `None`) format to interpret data
         sources. Will be inferred automatically if not specified.  Valid
         formats are `'auto'`, `'csv'`, `'excel'`, `'feather'`,
         `'fwf'`, `'hdf5'` (cache file produced during previous training),
         `'html'` (file containing a single HTML `<table>`), `'json'`, `'jsonl'`,
         `'parquet'`, `'pickle'` (pickled Pandas DataFrame), `'sas'`, `'spss'`,
         `'stata'`, `'tsv'`.
     :param split: (str, default: `full`) split on which
         to perform predictions. Valid values are `'training'`, `'validation'`,
         `'test'` and `'full'`.
     :param batch_size: (int, default `128`) size of batches for processing.
     :param skip_save_unprocessed_output: (bool, default: `False`) by default
         predictions and their probabilities are saved in both raw
         unprocessed numpy files containing tensors and as postprocessed
         CSV files (one for each output feature). If this parameter is True,
         only the CSV ones are saved and the numpy ones are skipped.
     :param skip_save_predictions: (bool, default: `False`) skips saving test
         predictions CSV files
     :param skip_save_eval_stats: (bool, default: `False`) skips saving test
         statistics JSON file
    :param skip_collect_predictions: (bool, default: `False`) skips
         collecting post-processed predictions during eval.
     :param skip_collect_overall_stats: (bool, default: `False`) skips
         collecting overall stats during eval.
     :param output_directory: (str, default: `'results'`) the directory that
         will contain the training statistics, TensorBoard logs, the saved
         model and the training progress files.
     :param gpus: (list, default: `None`) list of GPUs that are available
         for training.
     :param gpu_memory_limit: (float: default: `None`) maximum memory fraction
            [0, 1] allowed to allocate per GPU device.
     :param allow_parallel_threads: (bool, default: `True`) allow PyTorch
         to use multithreading parallelism to improve performance at
         the cost of determinism.
     :param callbacks: (list, default: `None`) a list of
         `ludwig.callbacks.Callback` objects that provide hooks into the
         Ludwig pipeline.
     :param backend: (Union[Backend, str]) `Backend` or string name
         of backend to use to execute preprocessing / training steps.
     :param logging_level: (int) Log level that will be sent to stderr.

     # Returns

     :return: (`None`)
    """
    model = LudwigModel.load(
        model_path,
        logging_level=logging_level,
        backend=backend,
        gpus=gpus,
        gpu_memory_limit=gpu_memory_limit,
        allow_parallel_threads=allow_parallel_threads,
        callbacks=callbacks,
    )
    model.evaluate(
        dataset=dataset,
        data_format=data_format,
        batch_size=batch_size,
        split=split,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
        skip_save_eval_stats=skip_save_eval_stats,
        collect_predictions=not skip_collect_predictions,
        collect_overall_stats=not skip_collect_overall_stats,
        output_directory=output_directory,
        return_type="dict",
    )


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model "
        "and evaluates its performance by comparing"
        "its predictions with ground truth.",
        prog="ludwig evaluate",
        usage="%(prog)s [options]",
    )

    # ---------------
    # Data parameters
    # ---------------
    parser.add_argument("--dataset", help="input data file path", required=True)
    parser.add_argument(
        "--data_format",
        help="format of the input data",
        default="auto",
        choices=[
            "auto",
            "csv",
            "excel",
            "feather",
            "fwf",
            "hdf5",
            "html" "tables",
            "json",
            "jsonl",
            "parquet",
            "pickle",
            "sas",
            "spss",
            "stata",
            "tsv",
        ],
    )
    parser.add_argument(
        "-s", "--split", default=FULL, choices=[TRAINING, VALIDATION, TEST, FULL], help="the split to test the model on"
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)

    # -------------------------
    # Output results parameters
    # -------------------------
    parser.add_argument(
        "-od", "--output_directory", type=str, default="results", help="directory that contains the results"
    )
    parser.add_argument(
        "-ssuo",
        "--skip_save_unprocessed_output",
        help="skips saving intermediate NPY output files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-sses",
        "--skip_save_eval_stats",
        help="skips saving intermediate JSON eval statistics",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-scp", "--skip_collect_predictions", help="skips collecting predictions", action="store_true", default=False
    )
    parser.add_argument(
        "-scos",
        "--skip_collect_overall_stats",
        help="skips collecting overall stats",
        action="store_true",
        default=False,
    )

    # ------------------
    # Generic parameters
    # ------------------
    parser.add_argument("-bs", "--batch_size", type=int, default=128, help="size of batches")

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument("-g", "--gpus", type=int, default=0, help="list of gpu to use")
    parser.add_argument(
        "-gml",
        "--gpu_memory_limit",
        type=float,
        default=None,
        help="maximum memory fraction [0, 1] allowed to allocate per GPU device",
    )
    parser.add_argument(
        "-dpt",
        "--disable_parallel_threads",
        action="store_false",
        dest="allow_parallel_threads",
        help="disable PyTorch from using multithreading for reproducibility",
    )
    parser.add_argument(
        "-b",
        "--backend",
        help="specifies backend to use for parallel / distributed execution, "
        "defaults to local execution or Horovod if called using horovodrun",
        choices=ALL_BACKENDS,
    )
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)
    args.evaluate_performance = True

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("evaluate", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.test_performance")

    args.backend = initialize_backend(args.backend)
    if args.backend.is_coordinator():
        print_ludwig("Evaluate", LUDWIG_VERSION)
        logger.info(f"Dataset path: {args.dataset}")
        logger.info(f"Model path: {args.model_path}")
        logger.info("")

    evaluate_cli(**vars(args))


if __name__ == "__main__":
    cli(sys.argv[1:])



#! /usr/bin/env python
# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os

import modin.pandas as pd
import numpy as np

from ludwig.data.dataframe.base import DataFrameEngine
from ludwig.globals import PREDICTIONS_SHAPES_FILE_NAME
from ludwig.utils.data_utils import get_pa_schema, load_json, save_json, split_by_slices
from ludwig.utils.dataframe_utils import flatten_df, unflatten_df


class ModinEngine(DataFrameEngine):
    def __init__(self, **kwargs):
        super().__init__()

    def df_like(self, df, proc_cols):
        # df argument unused for pandas, which can instantiate df directly
        return pd.DataFrame(proc_cols)

    def parallelize(self, data):
        return data

    def persist(self, data):
        return data

    def compute(self, data):
        return data

    def from_pandas(self, df):
        return pd.DataFrame(df)

    def map_objects(self, series, map_fn, meta=None):
        return series.map(map_fn)

    def map_batches(self, df, map_fn, enable_tensor_extension_casting=True):
        return map_fn(df)

    def map_partitions(self, series, map_fn, meta=None):
        return map_fn(series)

    def apply_objects(self, df, apply_fn, meta=None):
        return df.apply(apply_fn, axis=1)

    def reduce_objects(self, series, reduce_fn):
        return reduce_fn(series)

    def split(self, df, probabilities):
        return split_by_slices(df.iloc, len(df), probabilities)

    def remove_empty_partitions(self, df):
        return df

    def to_parquet(self, df, path, index=False):
        schema = get_pa_schema(df)
        df.to_parquet(
            path,
            engine="pyarrow",
            index=index,
            schema=schema,
        )

    def write_predictions(self, df: pd.DataFrame, path: str):
        df, column_shapes = flatten_df(df, self)
        self.to_parquet(df, path)
        save_json(os.path.join(os.path.dirname(path), PREDICTIONS_SHAPES_FILE_NAME), column_shapes)

    def read_predictions(self, path: str) -> pd.DataFrame:
        pred_df = pd.read_parquet(path)
        column_shapes = load_json(os.path.join(os.path.dirname(path), PREDICTIONS_SHAPES_FILE_NAME))
        return unflatten_df(pred_df, column_shapes, self)

    def to_ray_dataset(self, df):
        from ray.data import from_modin

        return from_modin(df)

    def from_ray_dataset(self, dataset) -> pd.DataFrame:
        return dataset.to_modin()

    def reset_index(self, df):
        return df.reset_index(drop=True)

    @property
    def array_lib(self):
        return np

    @property
    def df_lib(self):
        return pd

    @property
    def partitioned(self):
        return False

    def set_parallelism(self, parallelism):
        pass



import io
import os
import uuid
from unittest import mock

import pandas as pd
import pytest

import ludwig.datasets
from ludwig.api import LudwigModel
from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.loaders.dataset_loader import DatasetState

SUPPORTED_UNCOMPRESSED_FILETYPES = ["json", "jsonl", "tsv", "csv"]


def test_load_csv_dataset(tmpdir):
    input_df = pd.DataFrame(
        {"name": ["Raphael", "Donatello"], "mask": ["red", "purple"], "weapon": ["sai", "bo staff"], "split": [0, 1]}
    )

    extracted_filename = "input.csv"
    compression_opts = dict(method="zip", archive_name=extracted_filename)

    archive_filename = os.path.join(tmpdir, "archive.zip")
    input_df.to_csv(archive_filename, index=False, compression=compression_opts)

    config = DatasetConfig(
        version=1.0,
        name="fake_csv_dataset",
        download_urls=["file://" + archive_filename],
    )

    ludwig.datasets._get_dataset_configs.cache_clear()
    with mock.patch("ludwig.datasets._load_dataset_config", return_value=config):
        dataset = ludwig.datasets.get_dataset("fake_csv_dataset", cache_dir=tmpdir)

        assert not dataset.state == DatasetState.DOWNLOADED
        assert not dataset.state == DatasetState.TRANSFORMED

        output_df = dataset.load()
        pd.testing.assert_frame_equal(input_df, output_df)

        assert dataset.state == DatasetState.TRANSFORMED
    ludwig.datasets._get_dataset_configs.cache_clear()


@pytest.mark.parametrize("f_type", SUPPORTED_UNCOMPRESSED_FILETYPES)
def test_multifile_join_dataset(tmpdir, f_type):
    if f_type != "jsonl":
        train_df = pd.DataFrame(
            {"name": ["Raphael", "Donatello"], "mask": ["red", "purple"], "weapon": ["sai", "bo staff"]}
        )

        test_df = pd.DataFrame({"name": ["Jack", "Bob"], "mask": ["green", "yellow"], "weapon": ["knife", "gun"]})

        val_df = pd.DataFrame({"name": ["Tom"], "mask": ["pink"], "weapon": ["stick"]})
    else:
        train_df = pd.DataFrame([{"name": "joe"}, {"mask": "green"}, {"weapon": "stick"}])
        test_df = pd.DataFrame([{"name": "janice"}, {"mask": "black"}, {"weapon": "gun"}])
        val_df = pd.DataFrame([{"name": "sara"}, {"mask": "pink"}, {"weapon": "gun"}])

    # filetypes = ['json', 'tsv', 'jsonl']
    train_filename = "train." + f_type
    test_filename = "test." + f_type
    val_filename = "val." + f_type
    train_filepath = os.path.join(tmpdir, train_filename)
    test_filepath = os.path.join(tmpdir, test_filename)
    val_filepath = os.path.join(tmpdir, val_filename)

    if f_type == "json":
        train_df.to_json(train_filepath)
        test_df.to_json(test_filepath)
        val_df.to_json(val_filepath)
    elif f_type == "jsonl":
        train_df.to_json(train_filepath, orient="records", lines=True)
        test_df.to_json(test_filepath, orient="records", lines=True)
        val_df.to_json(val_filepath, orient="records", lines=True)
    elif f_type == "tsv":
        train_df.to_csv(train_filepath, sep="\t")
        test_df.to_csv(test_filepath, sep="\t")
        val_df.to_csv(val_filepath, sep="\t")
    else:
        train_df.to_csv(train_filepath)
        test_df.to_csv(test_filepath)
        val_df.to_csv(val_filepath)

    config = DatasetConfig(
        version=1.0,
        name="fake_multifile_dataset",
        download_urls=["file://" + train_filepath, "file://" + test_filepath, "file://" + val_filepath],
        train_filenames=train_filename,
        validation_filenames=val_filename,
        test_filenames=test_filename,
    )

    ludwig.datasets._get_dataset_configs.cache_clear()
    with mock.patch("ludwig.datasets._load_dataset_config", return_value=config):
        dataset = ludwig.datasets.get_dataset("fake_multifile_dataset", cache_dir=tmpdir)

        assert not dataset.state == DatasetState.DOWNLOADED
        assert not dataset.state == DatasetState.TRANSFORMED

        output_df = dataset.load()
        assert output_df.shape[0] == train_df.shape[0] + test_df.shape[0] + val_df.shape[0]

        assert dataset.state == DatasetState.TRANSFORMED
    ludwig.datasets._get_dataset_configs.cache_clear()


def test_get_datasets_info():
    dataset_output_features = ludwig.datasets.get_datasets_output_features()
    assert len(dataset_output_features) > 1
    assert isinstance(dataset_output_features, dict)
    assert dataset_output_features["titanic"].get("name", None)
    assert dataset_output_features["titanic"].get("output_features", None)
    assert isinstance(dataset_output_features["titanic"]["output_features"], list)
    assert dataset_output_features["titanic"]["output_features"][0].get("name", None)
    assert dataset_output_features["titanic"]["output_features"][0].get("type", None)

    dataset_output_features = ludwig.datasets.get_datasets_output_features(dataset="titanic")
    assert len(dataset_output_features["output_features"]) == 1
    assert dataset_output_features["name"] == "titanic"


def test_get_dataset_buffer():
    buffer = ludwig.datasets.get_buffer("iris")

    assert isinstance(buffer, io.BytesIO)


def test_preprocess_dataset_uri(tmpdir):
    input_df = pd.DataFrame(
        {
            "input": ["a", "b", "a", "b", "a", "b", "c", "c", "a", "b"],
            "output": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "split": [0, 0, 0, 0, 0, 0, 0, 1, 2, 2],
        }
    )

    extracted_filename = "input.csv"
    compression_opts = dict(method="zip", archive_name=extracted_filename)

    archive_filename = os.path.join(tmpdir, "archive.zip")
    input_df.to_csv(archive_filename, index=False, compression=compression_opts)

    dataset_name = f"fake_csv_dataset_{uuid.uuid4().hex}"
    config = DatasetConfig(
        version=1.0,
        name=dataset_name,
        download_urls=["file://" + archive_filename],
    )

    model_config = {
        "input_features": [{"name": "input", "type": "category"}],
        "output_features": [{"name": "output", "type": "number"}],
        "preprocessing": {"split": {"type": "fixed"}},
    }

    ludwig.datasets._get_dataset_configs.cache_clear()
    with mock.patch("ludwig.datasets._load_dataset_config", return_value=config):
        with mock.patch("ludwig.datasets.loaders.dataset_loader.get_default_cache_location", return_value=str(tmpdir)):
            model = LudwigModel(model_config, backend="local")

            proc_result = model.preprocess(dataset=f"ludwig://{dataset_name}")
            train_df1 = proc_result.training_set.to_df()
            val_df1 = proc_result.validation_set.to_df()
            test_df1 = proc_result.test_set.to_df()

            assert len(train_df1) == 7
            assert len(val_df1) == 1
            assert len(test_df1) == 2

            proc_result_split = model.preprocess(
                training_set=f"ludwig://{dataset_name}",
                validation_set=f"ludwig://{dataset_name}",
                test_set=f"ludwig://{dataset_name}",
            )
            train_df2 = proc_result_split.training_set.to_df()
            val_df2 = proc_result_split.validation_set.to_df()
            test_df2 = proc_result_split.test_set.to_df()

            assert len(train_df2) == 7
            assert len(val_df2) == 1
            assert len(test_df2) == 2

            assert train_df1.equals(train_df2)
            assert val_df1.equals(val_df2)
            assert test_df1.equals(test_df2)

    ludwig.datasets._get_dataset_configs.cache_clear()



from typing import List

import pytest
import torch

from ludwig.encoders.set_encoders import SetSparseEncoder
from ludwig.utils.misc_utils import set_random_seed
from ludwig.utils.torch_utils import get_torch_device
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated

RANDOM_SEED = 1919
DEVICE = get_torch_device()


@pytest.mark.parametrize("num_fc_layers", [0, 2])
@pytest.mark.parametrize("vocab", [["a", "b", "c", "d", "e", "f", "g", "h"]])
@pytest.mark.parametrize("embedding_size", [10])
@pytest.mark.parametrize("representation", ["sparse"])
def test_set_encoder(
    vocab: List[str],
    embedding_size: int,
    representation: str,
    num_fc_layers: int,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    # setup encoder to test
    set_encoder = SetSparseEncoder(
        vocab=vocab,
        representation=representation,
        embedding_size=embedding_size,
        num_fc_layers=num_fc_layers,
    ).to(DEVICE)
    inputs = torch.randint(0, 2, size=(2, len(vocab))).bool().to(DEVICE)
    outputs = set_encoder(inputs)
    assert outputs.shape[1:] == set_encoder.output_shape

    # check for parameter updating
    target = torch.randn(outputs.shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(set_encoder, (inputs,), target)
    assert tpc == upc, f"Failed to update parameters. Parameters not updated: {not_updated}"



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import logging
import os
import sys
from typing import List, Optional, Union

import pandas as pd

from ludwig.api import kfold_cross_validate, LudwigModel
from ludwig.backend import ALL_BACKENDS, Backend, initialize_backend
from ludwig.callbacks import Callback
from ludwig.constants import CONTINUE_PROMPT, FULL, HYPEROPT, HYPEROPT_WARNING, TEST, TRAINING, VALIDATION
from ludwig.contrib import add_contrib_callback_args
from ludwig.globals import LUDWIG_VERSION
from ludwig.utils.data_utils import load_config_from_str, load_yaml, save_json
from ludwig.utils.defaults import default_random_seed
from ludwig.utils.print_utils import get_logging_level_registry, print_ludwig, query_yes_no

logger = logging.getLogger(__name__)


def experiment_cli(
    config: Union[str, dict],
    dataset: Union[str, dict, pd.DataFrame] = None,
    training_set: Union[str, dict, pd.DataFrame] = None,
    validation_set: Union[str, dict, pd.DataFrame] = None,
    test_set: Union[str, dict, pd.DataFrame] = None,
    training_set_metadata: Union[str, dict] = None,
    data_format: str = None,
    experiment_name: str = "experiment",
    model_name: str = "run",
    model_load_path: str = None,
    model_resume_path: str = None,
    eval_split: str = TEST,
    skip_save_training_description: bool = False,
    skip_save_training_statistics: bool = False,
    skip_save_model: bool = False,
    skip_save_progress: bool = False,
    skip_save_log: bool = False,
    skip_save_processed_input: bool = False,
    skip_save_unprocessed_output: bool = False,
    skip_save_predictions: bool = False,
    skip_save_eval_stats: bool = False,
    skip_collect_predictions: bool = False,
    skip_collect_overall_stats: bool = False,
    output_directory: str = "results",
    gpus: Union[str, int, List[int]] = None,
    gpu_memory_limit: Optional[float] = None,
    allow_parallel_threads: bool = True,
    callbacks: List[Callback] = None,
    backend: Union[Backend, str] = None,
    random_seed: int = default_random_seed,
    logging_level: int = logging.INFO,
    **kwargs,
):
    """Trains a model on a dataset's training and validation splits and uses it to predict on the test split. It
    saves the trained model and the statistics of training and testing.

     # Inputs

     :param config: (Union[str, dict]) in-memory representation of
             config or string path to a YAML config file.
     :param dataset: (Union[str, dict, pandas.DataFrame], default: `None`)
         source containing the entire dataset to be used in the experiment.
         If it has a split column, it will be used for splitting (0 for train,
         1 for validation, 2 for test), otherwise the dataset will be
         randomly split.
     :param training_set: (Union[str, dict, pandas.DataFrame], default: `None`)
         source containing training data.
     :param validation_set: (Union[str, dict, pandas.DataFrame], default: `None`)
         source containing validation data.
     :param test_set: (Union[str, dict, pandas.DataFrame], default: `None`)
         source containing test data.
     :param training_set_metadata: (Union[str, dict], default: `None`)
         metadata JSON file or loaded metadata.  Intermediate preprocessed
         structure containing the mappings of the input
         dataset created the first time an input file is used in the same
         directory with the same name and a '.meta.json' extension.
     :param data_format: (str, default: `None`) format to interpret data
         sources. Will be inferred automatically if not specified.  Valid
         formats are `'auto'`, `'csv'`, `'excel'`, `'feather'`,
         `'fwf'`, `'hdf5'` (cache file produced during previous training),
         `'html'` (file containing a single HTML `<table>`), `'json'`, `'jsonl'`,
         `'parquet'`, `'pickle'` (pickled Pandas DataFrame), `'sas'`, `'spss'`,
         `'stata'`, `'tsv'`.
     :param experiment_name: (str, default: `'experiment'`) name for
         the experiment.
     :param model_name: (str, default: `'run'`) name of the model that is
         being used.
     :param model_load_path: (str, default: `None`) if this is specified the
         loaded model will be used as initialization
         (useful for transfer learning).
     :param model_resume_path: (str, default: `None`) resumes training of
         the model from the path specified. The config is restored.
         In addition to config, training statistics and loss for
         epoch and the state of the optimizer are restored such that
         training can be effectively continued from a previously interrupted
         training process.
     :param eval_split: (str, default: `test`) split on which
         to perform evaluation. Valid values are `training`, `validation`
         and `test`.
     :param skip_save_training_description: (bool, default: `False`) disables
         saving the description JSON file.
     :param skip_save_training_statistics: (bool, default: `False`) disables
         saving training statistics JSON file.
     :param skip_save_model: (bool, default: `False`) disables
         saving model weights and hyperparameters each time the model
         improves. By default Ludwig saves model weights after each epoch
         the validation metric improves, but if the model is really big
         that can be time consuming. If you do not want to keep
         the weights and just find out what performance a model can get
         with a set of hyperparameters, use this parameter to skip it,
         but the model will not be loadable later on and the returned model
         will have the weights obtained at the end of training, instead of
         the weights of the epoch with the best validation performance.
    :param skip_save_progress: (bool, default: `False`) disables saving
         progress each epoch. By default Ludwig saves weights and stats
         after each epoch for enabling resuming of training, but if
         the model is really big that can be time consuming and will uses
         twice as much space, use this parameter to skip it, but training
         cannot be resumed later on.
     :param skip_save_log: (bool, default: `False`) disables saving
         TensorBoard logs. By default Ludwig saves logs for the TensorBoard,
         but if it is not needed turning it off can slightly increase the
         overall speed.
     :param skip_save_processed_input: (bool, default: `False`) if input
         dataset is provided it is preprocessed and cached by saving an HDF5
         and JSON files to avoid running the preprocessing again. If this
         parameter is `False`, the HDF5 and JSON file are not saved.
     :param skip_save_unprocessed_output: (bool, default: `False`) by default
         predictions and their probabilities are saved in both raw
         unprocessed numpy files containing tensors and as postprocessed
         CSV files (one for each output feature). If this parameter is True,
         only the CSV ones are saved and the numpy ones are skipped.
     :param skip_save_predictions: (bool, default: `False`) skips saving test
         predictions CSV files
     :param skip_save_eval_stats: (bool, default: `False`) skips saving test
         statistics JSON file
    :param skip_collect_predictions: (bool, default: `False`) skips
         collecting post-processed predictions during eval.
     :param skip_collect_overall_stats: (bool, default: `False`) skips
         collecting overall stats during eval.
     :param output_directory: (str, default: `'results'`) the directory that
         will contain the training statistics, TensorBoard logs, the saved
         model and the training progress files.
     :param gpus: (list, default: `None`) list of GPUs that are available
         for training.
     :param gpu_memory_limit: (float: default: `None`) maximum memory fraction
            [0, 1] allowed to allocate per GPU device.
     :param allow_parallel_threads: (bool, default: `True`) allow PyTorch
         to use multithreading parallelism to improve performance at
         the cost of determinism.
     :param callbacks: (list, default: `None`) a list of
         `ludwig.callbacks.Callback` objects that provide hooks into the
         Ludwig pipeline.
     :param backend: (Union[Backend, str]) `Backend` or string name
         of backend to use to execute preprocessing / training steps.
     :param random_seed: (int: default: 42) random seed used for weights
         initialization, splits and any other random function.
     :param logging_level: (int) Log level that will be sent to stderr.

     # Return
     :return: (Tuple[LudwigModel, dict, dict, tuple, str)):
        `(model, evaluation_statistics, training_statistics, preprocessed_data, output_directory)`
         `model` LudwigModel instance
         `evaluation_statistics` dictionary with evaluation performance
             statistics on the test_set,
         `training_statistics` is a nested dictionary of dataset -> feature_name -> metric_name -> List of metrics.
                Each metric corresponds to each training checkpoint.
         `preprocessed_data` tuple containing preprocessed
         `(training_set, validation_set, test_set)`, `output_directory`
         filepath string to where results are stored.
    """
    if HYPEROPT in config:
        if not query_yes_no(HYPEROPT_WARNING + CONTINUE_PROMPT):
            exit(1)

    if isinstance(config, str):
        config = load_yaml(config)
    backend = initialize_backend(backend or config.get("backend"))

    if model_load_path:
        model = LudwigModel.load(
            model_load_path,
            logging_level=logging_level,
            backend=backend,
            gpus=gpus,
            gpu_memory_limit=gpu_memory_limit,
            allow_parallel_threads=allow_parallel_threads,
            callbacks=callbacks,
        )
    else:
        model = LudwigModel(
            config=config,
            logging_level=logging_level,
            backend=backend,
            gpus=gpus,
            gpu_memory_limit=gpu_memory_limit,
            allow_parallel_threads=allow_parallel_threads,
            callbacks=callbacks,
        )
    (eval_stats, train_stats, preprocessed_data, output_directory) = model.experiment(
        dataset=dataset,
        training_set=training_set,
        validation_set=validation_set,
        test_set=test_set,
        training_set_metadata=training_set_metadata,
        data_format=data_format,
        experiment_name=experiment_name,
        model_name=model_name,
        model_resume_path=model_resume_path,
        eval_split=eval_split,
        skip_save_training_description=skip_save_training_description,
        skip_save_training_statistics=skip_save_training_statistics,
        skip_save_model=skip_save_model,
        skip_save_progress=skip_save_progress,
        skip_save_log=skip_save_log,
        skip_save_processed_input=skip_save_processed_input,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
        skip_save_eval_stats=skip_save_eval_stats,
        skip_collect_predictions=skip_collect_predictions,
        skip_collect_overall_stats=skip_collect_overall_stats,
        output_directory=output_directory,
        random_seed=random_seed,
    )

    return model, eval_stats, train_stats, preprocessed_data, output_directory


def kfold_cross_validate_cli(
    k_fold,
    config=None,
    dataset=None,
    data_format=None,
    output_directory="results",
    random_seed=default_random_seed,
    skip_save_k_fold_split_indices=False,
    **kwargs,
):
    """Wrapper function to performs k-fold cross validation.

    # Inputs
    :param k_fold: (int) number of folds to create for the cross-validation
    :param config: (Union[str, dict], default: None) a dictionary or file path
            containing model configuration. Refer to the [User Guide]
           (http://ludwig.ai/user_guide/#model-config) for details.
    :param dataset: (string, default: None)
    :param output_directory: (string, default: 'results')
    :param random_seed: (int) Random seed used k-fold splits.
    :param skip_save_k_fold_split_indices: (boolean, default: False) Disables
            saving k-fold split indices

    :return: None
    """

    (kfold_cv_stats, kfold_split_indices) = kfold_cross_validate(
        k_fold,
        config=config,
        dataset=dataset,
        data_format=data_format,
        output_directory=output_directory,
        random_seed=random_seed,
    )

    # save k-fold cv statistics
    save_json(os.path.join(output_directory, "kfold_training_statistics.json"), kfold_cv_stats)

    # save k-fold split indices
    if not skip_save_k_fold_split_indices:
        save_json(os.path.join(output_directory, "kfold_split_indices.json"), kfold_split_indices)


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script trains and evaluates a model", prog="ludwig experiment", usage="%(prog)s [options]"
    )

    # ----------------------------
    # Experiment naming parameters
    # ----------------------------
    parser.add_argument("--output_directory", type=str, default="results", help="directory that contains the results")
    parser.add_argument("--experiment_name", type=str, default="experiment", help="experiment name")
    parser.add_argument("--model_name", type=str, default="run", help="name for the model")

    # ---------------
    # Data parameters
    # ---------------
    parser.add_argument(
        "--dataset",
        help="input data file path. "
        "If it has a split column, it will be used for splitting "
        "(0: train, 1: validation, 2: test), "
        "otherwise the dataset will be randomly split",
    )
    parser.add_argument("--training_set", help="input train data file path")
    parser.add_argument("--validation_set", help="input validation data file path")
    parser.add_argument("--test_set", help="input test data file path")

    parser.add_argument(
        "--training_set_metadata",
        help="input metadata JSON file path. An intermediate preprocessed file "
        "containing the mappings of the input file created "
        "the first time a file is used, in the same directory "
        "with the same name and a .json extension",
    )

    parser.add_argument(
        "--data_format",
        help="format of the input data",
        default="auto",
        choices=[
            "auto",
            "csv",
            "excel",
            "feather",
            "fwf",
            "hdf5",
            "html" "tables",
            "json",
            "jsonl",
            "parquet",
            "pickle",
            "sas",
            "spss",
            "stata",
            "tsv",
        ],
    )

    parser.add_argument(
        "-es",
        "--eval_split",
        default=TEST,
        choices=[TRAINING, VALIDATION, TEST, FULL],
        help="the split to evaluate the model on",
    )

    parser.add_argument(
        "-sspi",
        "--skip_save_processed_input",
        help="skips saving intermediate HDF5 and JSON files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-ssuo",
        "--skip_save_unprocessed_output",
        help="skips saving intermediate NPY output files",
        action="store_true",
        default=False,
    )

    # -----------------
    # K-fold parameters
    # -----------------
    parser.add_argument(
        "-kf", "--k_fold", type=int, default=None, help="number of folds for a k-fold cross validation run "
    )
    parser.add_argument(
        "-skfsi",
        "--skip_save_k_fold_split_indices",
        action="store_true",
        default=False,
        help="disables saving indices generated to split training data set "
        "for the k-fold cross validation run, but if it is not needed "
        "turning it off can slightly increase the overall speed",
    )

    # ----------------
    # Model parameters
    # ----------------
    config = parser.add_mutually_exclusive_group(required=True)
    config.add_argument(
        "-c",
        "--config",
        type=load_yaml,
        help="Path to the YAML file containing the model configuration",
    )
    config.add_argument(
        "-cs",
        "--config_str",
        dest="config",
        type=load_config_from_str,
        help="JSON or YAML serialized string of the model configuration",
    )

    parser.add_argument("-mlp", "--model_load_path", help="path of a pretrained model to load as initialization")
    parser.add_argument("-mrp", "--model_resume_path", help="path of the model directory to resume training of")
    parser.add_argument(
        "-sstd",
        "--skip_save_training_description",
        action="store_true",
        default=False,
        help="disables saving the description JSON file",
    )
    parser.add_argument(
        "-ssts",
        "--skip_save_training_statistics",
        action="store_true",
        default=False,
        help="disables saving training statistics JSON file",
    )
    parser.add_argument(
        "-sstp",
        "--skip_save_predictions",
        help="skips saving test predictions CSV files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-sstes",
        "--skip_save_eval_stats",
        help="skips saving eval statistics JSON file",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-ssm",
        "--skip_save_model",
        action="store_true",
        default=False,
        help="disables saving model weights and hyperparameters each time "
        "the model improves. "
        "By default Ludwig saves model weights after each epoch "
        "the validation metric improves, but if the model is really big "
        "that can be time consuming. If you do not want to keep "
        "the weights and just find out what performance a model can get "
        "with a set of hyperparameters, use this parameter to skip it,"
        "but the model will not be loadable later on",
    )
    parser.add_argument(
        "-ssp",
        "--skip_save_progress",
        action="store_true",
        default=False,
        help="disables saving progress each epoch. By default Ludwig saves "
        "weights and stats after each epoch for enabling resuming "
        "of training, but if the model is really big that can be "
        "time consuming and will uses twice as much space, use "
        "this parameter to skip it, but training cannot be resumed "
        "later on",
    )
    parser.add_argument(
        "-ssl",
        "--skip_save_log",
        action="store_true",
        default=False,
        help="disables saving TensorBoard logs. By default Ludwig saves "
        "logs for the TensorBoard, but if it is not needed turning it off "
        "can slightly increase the overall speed",
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        "-rs",
        "--random_seed",
        type=int,
        default=42,
        help="a random seed that is going to be used anywhere there is a call "
        "to a random number generator: data splitting, parameter "
        "initialization and training set shuffling",
    )
    parser.add_argument("-g", "--gpus", nargs="+", type=int, default=None, help="list of GPUs to use")
    parser.add_argument(
        "-gml",
        "--gpu_memory_limit",
        type=float,
        default=None,
        help="maximum memory fraction [0, 1] allowed to allocate per GPU device",
    )
    parser.add_argument(
        "-dpt",
        "--disable_parallel_threads",
        action="store_false",
        dest="allow_parallel_threads",
        help="disable PyTorch from using multithreading for reproducibility",
    )
    parser.add_argument(
        "-b",
        "--backend",
        help="specifies backend to use for parallel / distributed execution, "
        "defaults to local execution or Horovod if called using horovodrun",
        choices=ALL_BACKENDS,
    )
    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("experiment", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.experiment")

    args.backend = initialize_backend(args.backend or args.config.get("backend"))
    if args.backend.is_coordinator():
        print_ludwig("Experiment", LUDWIG_VERSION)

    if args.k_fold is None:
        experiment_cli(**vars(args))
    else:
        kfold_cross_validate_cli(**vars(args))


if __name__ == "__main__":
    cli(sys.argv[1:])



import copy
from dataclasses import field
from typing import Any, Dict, List, Optional, Union

from marshmallow import fields, ValidationError

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import TYPE
from ludwig.schema import utils as schema_utils
from ludwig.schema.features.augmentation.base import BaseAugmentationConfig
from ludwig.utils.registry import Registry

_augmentation_config_registry = Registry()


@DeveloperAPI
def get_augmentation_config_registry() -> Registry:
    return _augmentation_config_registry


@DeveloperAPI
def register_augmentation_config(name: str, features: Union[str, List[str]]):
    if isinstance(features, str):
        features = [features]

    def wrap(cls):
        for feature in features:
            augmentation_registry = get_augmentation_config_registry().get(feature, {})
            augmentation_registry[name] = cls
            get_augmentation_config_registry()[feature] = augmentation_registry
        return cls

    return wrap


@DeveloperAPI
def get_augmentation_cls(feature: str, name: str):
    return get_augmentation_config_registry()[feature][name]


@DeveloperAPI
def get_augmentation_classes(feature: str):
    return get_augmentation_config_registry()[feature]


@DeveloperAPI
def AugmentationDataclassField(
    feature_type: str,
    default: Union[str, BaseAugmentationConfig] = False,
    default_augmentations: Optional[List[BaseAugmentationConfig]] = None,
    description: str = "",
):
    """Custom dataclass field that when used inside a dataclass will allow the user to specify an augmentation
    config.

    Args:
        default: The default augmentation config to use.
        default_augmentations: The default list of augmentations to use when param value is set to `True`.
        description: The description of the augmentation config.

    Returns: Initialized dataclass field that converts a list with params to an augmentation config.
    """

    default_augmentations = default_augmentations or []
    default_augmentations = [a.to_dict() for a in default_augmentations]

    if isinstance(default, bool):
        default = default_augmentations if default else []

    class AugmentationContainerMarshmallowField(fields.Field):
        """Custom marshmallow field that deserializes a list for a valid augmentation config from the
        augmentation_registry and creates a corresponding JSON schema for external usage."""

        def _deserialize(self, value, attr, data, **kwargs):
            if isinstance(value, bool):
                value = default_augmentations if value else []

            if not isinstance(value, list):
                raise ValidationError(f"Augmentation config must be a list, found: {type(value)}")

            augmentation_classes = get_augmentation_classes(feature_type)
            augmentation_list = []
            for augmentation in value:
                augmentation_op = augmentation[TYPE]
                if augmentation_op in augmentation_classes:
                    augmentation_cls = augmentation_classes[augmentation_op]
                    pre = augmentation_cls()
                    try:
                        augmentation_list.append(pre.Schema().load(augmentation))
                    except (TypeError, ValidationError) as error:
                        raise ValidationError(
                            f"Invalid augmentation params: {value}, see `{pre}` definition. Error: {error}"
                        )
                else:
                    raise ValidationError(
                        f"Invalid augmentation type: '{augmentation_op}', "
                        f"expected one of: {list(augmentation_classes.keys())}"
                    )
            return augmentation_list

        @staticmethod
        def _jsonschema_type_mapping():
            return get_augmentation_list_jsonschema(feature_type, default)

    try:
        assert isinstance(default, list), "Augmentation config must be a list."
        load_augmentation_list = []
        dump_augmentation_list = []
        for augmentation in default:
            augmentation_op = augmentation[TYPE]
            augmentation_cls = get_augmentation_cls(feature_type, augmentation_op)
            pre = augmentation_cls()
            try:
                load_augmentation_list.append(pre.Schema().load(augmentation))
                dump_augmentation_list.append(pre.Schema().dump(augmentation))
            except (TypeError, ValidationError) as error:
                raise ValidationError(f"Invalid augmentation params: {default}, see `{pre}` definition. Error: {error}")

        load_default = lambda: copy.deepcopy(load_augmentation_list)
        dump_default = dump_augmentation_list

        return field(
            metadata={
                "marshmallow_field": AugmentationContainerMarshmallowField(
                    allow_none=False,
                    dump_default=dump_default,
                    load_default=load_default,
                )
            },
            default_factory=load_default,
        )
    except Exception as e:
        raise ValidationError(f"Unsupported augmentation type. See augmentation_registry. " f"Details: {e}")


@DeveloperAPI
def get_augmentation_list_jsonschema(feature_type: str, default: List[Dict[str, Any]]):
    """This function returns a JSON augmentation schema.

    Returns: JSON Schema
    """
    augmentation_types = sorted(list(get_augmentation_config_registry()[feature_type].keys()))
    schema = {
        "oneOf": [
            {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "type": {
                            "type": "string",
                            "enum": augmentation_types,
                            "title": "type",
                            "description": "Type of augmentation to apply.",
                        },
                    },
                    "additionalProperties": True,
                    "allOf": get_augmentation_list_conds(feature_type),
                    "required": ["type"],
                    "title": "augmentation",
                },
            },
            {
                "type": "boolean",
                "description": "Apply standard augmentation pipeline.",
            },
        ],
    }

    return schema


@DeveloperAPI
def get_augmentation_list_conds(feature_type: str):
    """This function returns a list of if-then JSON clauses for each augmentation type along with their properties
    and constraints.

    Returns: List of JSON clauses
    """
    conds = []
    for augmentation_op in get_augmentation_classes(feature_type):
        schema_cls = get_augmentation_cls(feature_type, augmentation_op)
        augmentation_schema = schema_utils.unload_jsonschema_from_marshmallow_class(schema_cls)
        augmentation_props = augmentation_schema["properties"]
        schema_utils.remove_duplicate_fields(augmentation_props)
        augmentation_cond = schema_utils.create_cond({"type": augmentation_op}, augmentation_props)
        conds.append(augmentation_cond)
    return conds



from copy import deepcopy
from datetime import datetime

import pytest
import torch

from ludwig.features import date_feature
from ludwig.features.date_feature import DateInputFeature
from ludwig.schema.features.date_feature import DateInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from ludwig.types import FeatureConfigDict
from ludwig.utils.misc_utils import merge_dict
from ludwig.utils.torch_utils import get_torch_device

BATCH_SIZE = 2
DATE_W_SIZE = 9
DEVICE = get_torch_device()


@pytest.fixture(scope="module")
def date_config():
    return {"name": "date_column_name", "type": "date"}


def test_date_input_feature(date_config: FeatureConfigDict):
    # setup image input feature definition
    feature_def = deepcopy(date_config)

    # pickup any other missing parameters
    defaults = DateInputFeatureConfig().to_dict()
    set_def = merge_dict(defaults, feature_def)

    # ensure no exceptions raised during build
    feature_config, _ = load_config_with_kwargs(DateInputFeatureConfig, set_def)
    input_feature_obj = DateInputFeature(feature_config).to(DEVICE)

    # check one forward pass through input feature
    input_tensor = input_feature_obj.create_sample_input(batch_size=BATCH_SIZE)
    assert input_tensor.shape == torch.Size((BATCH_SIZE, DATE_W_SIZE))
    assert input_tensor.dtype == torch.int32

    encoder_output = input_feature_obj(input_tensor)
    assert encoder_output["encoder_output"].shape == (BATCH_SIZE, *input_feature_obj.output_shape)


@pytest.mark.parametrize(
    "date_str,datetime_format,expected_list",
    [
        ("2012-02-26T13:51:50.417-07:00", None, [2012, 2, 26, 6, 57, 13, 51, 50, 49910]),
        ("2022-06-25 09:30:59", None, [2022, 6, 25, 5, 176, 9, 30, 59, 34259]),
        ("2022-06-25", None, [2022, 6, 25, 5, 176, 0, 0, 0, 0]),
    ],
)
def test_date_to_list(date_str, datetime_format, expected_list):
    preprocessing_parameters = None
    assert (
        date_feature.DateInputFeature.date_to_list(date_str, datetime_format, preprocessing_parameters) == expected_list
    )


def test_date_to_list__DatetimeObjectFromParsedJSON():
    preprocessing_parameters = None
    datetime_obj = datetime.fromisoformat("2022-06-25")
    assert date_feature.DateInputFeature.date_to_list(datetime_obj, None, preprocessing_parameters) == [
        2022,
        6,
        25,
        5,
        176,
        0,
        0,
        0,
        0,
    ]


def test_date_to_list__UsesFillValueOnInvalidDate():
    preprocessing_parameters = {"fill_value": "2013-02-26"}
    invalid_date_str = "2012abc-02"
    datetime_format = None
    assert date_feature.DateInputFeature.date_to_list(invalid_date_str, datetime_format, preprocessing_parameters) == [
        2013,
        2,
        26,
        1,
        57,
        0,
        0,
        0,
        0,
    ]



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import time
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)


class WithTimer:
    def __init__(self, title="", quiet=False):
        self.title = title
        self.quiet = quiet

    def elapsed(self):
        return time.time() - self.wall, time.process_time() - self.proc

    def enter(self):
        """Manually trigger enter."""
        self.__enter__()

    def __enter__(self):
        self.proc = time.process_time()
        self.wall = time.time()
        return self

    def __exit__(self, *args):
        if not self.quiet:
            elapsed_wp = self.elapsed()
            logger.info(f"Elapsed {self.title}: wall {elapsed_wp[0]:.06f}, sys {elapsed_wp[1]:.06f}")


class Timer:
    def __init__(self):
        self.reset()

    def reset(self):
        self._proc = time.process_time()
        self._wall = time.time()

    def elapsed(self):
        return self.wall(), self.proc()

    def elapsed_str(self):
        return strdelta(self.wall() * 1000.0), strdelta(self.proc() * 1000.0)

    def wall(self):
        return time.time() - self._wall

    def proc(self):
        return time.process_time() - self._proc

    def tic(self):
        """Like Matlab tic/toc for wall time and processor time."""
        self.reset()

    def toc(self):
        """Like Matlab tic/toc for wall time."""
        return self.wall()

    def tocproc(self):
        """Like Matlab tic/toc, but for processor time."""
        return self.proc()


def timestamp():
    return f"{datetime.now():%Y_%m_%d_%H_%M_%S}"


def strdelta(tdelta):
    if isinstance(tdelta, (int, float)):
        tdelta = timedelta(milliseconds=tdelta)
    d = {"D": tdelta.days}
    d["H"], rem = divmod(tdelta.seconds, 3600)
    d["M"], d["S"] = divmod(rem, 60)
    d["f"] = str(tdelta.microseconds)[0:4]
    if d["D"] > 0:
        t = "{D}d {H}h {M}m {S}.{f}s"
    elif d["H"] > 0:
        t = "{H}h {M}m {S}.{f}s"
    elif d["M"] > 0:
        t = "{M}m {S}.{f}s"
    else:
        t = "{S}.{f}s"
    return t.format(**d)



from typing import Optional

import pytest
import torch

from ludwig.modules.tabnet_modules import AttentiveTransformer, FeatureBlock, FeatureTransformer, TabNet
from ludwig.utils.entmax import sparsemax
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated

RANDOM_SEED = 67


@pytest.mark.parametrize(
    "input_tensor",
    [
        torch.tensor([[-1.0, 0.0, 1.0], [5.01, 4.0, -2.0]], dtype=torch.float32),
        torch.tensor(
            [[1.36762051e8, -1.36762051e8, 1.59594639e20], [1.59594639e37, 1.36762051e7, 1.26e6]], dtype=torch.float32
        ),
    ],
)
def test_sparsemax(input_tensor: torch.Tensor) -> None:
    output_tensor = sparsemax(input_tensor)

    assert isinstance(output_tensor, torch.Tensor)
    assert output_tensor.equal(torch.tensor([[0, 0, 1], [1, 0, 0]], dtype=torch.float32))


@pytest.mark.parametrize("bn_virtual_bs", [None, 7])
@pytest.mark.parametrize("external_shared_fc_layer", [True, False])
@pytest.mark.parametrize("apply_glu", [True, False])
@pytest.mark.parametrize("size", [4, 12])
@pytest.mark.parametrize("input_size", [2, 6])
@pytest.mark.parametrize("batch_size", [1, 16])
def test_feature_block(
    input_size,
    size: int,
    apply_glu: bool,
    external_shared_fc_layer: bool,
    bn_virtual_bs: Optional[int],
    batch_size: int,
) -> None:
    # setup synthetic tensor
    torch.manual_seed(RANDOM_SEED)
    input_tensor = torch.randn([batch_size, input_size], dtype=torch.float32)

    if external_shared_fc_layer:
        shared_fc_layer = torch.nn.Linear(input_size, size * 2 if apply_glu else size, bias=False)
    else:
        shared_fc_layer = None

    feature_block = FeatureBlock(
        input_size, size, apply_glu=apply_glu, shared_fc_layer=shared_fc_layer, bn_virtual_bs=bn_virtual_bs
    )

    output_tensor = feature_block(input_tensor)

    # check for expected structure and properties
    assert isinstance(output_tensor, torch.Tensor)
    assert output_tensor.shape == (batch_size, size)

    assert feature_block.input_shape[-1] == input_size
    assert feature_block.output_shape[-1] == size
    assert feature_block.input_dtype == torch.float32


@pytest.mark.parametrize("num_total_blocks, num_shared_blocks", [(4, 2), (6, 4), (3, 1)])
@pytest.mark.parametrize("virtual_batch_size", [None, 7])
@pytest.mark.parametrize("size", [4, 12])
@pytest.mark.parametrize("input_size", [2, 6])
@pytest.mark.parametrize("batch_size", [1, 16])
def test_feature_transformer(
    input_size: int,
    size: int,
    virtual_batch_size: Optional[int],
    num_total_blocks: int,
    num_shared_blocks: int,
    batch_size: int,
) -> None:
    # setup synthetic tensor
    torch.manual_seed(RANDOM_SEED)
    input_tensor = torch.randn([batch_size, input_size], dtype=torch.float32)

    feature_transformer = FeatureTransformer(
        input_size,
        size,
        bn_virtual_bs=virtual_batch_size,
        num_total_blocks=num_total_blocks,
        num_shared_blocks=num_shared_blocks,
    )

    output_tensor = feature_transformer(input_tensor)

    # check for expected structure and properties
    assert isinstance(output_tensor, torch.Tensor)
    assert output_tensor.shape == (batch_size, size)

    assert feature_transformer.input_shape[-1] == input_size
    assert feature_transformer.output_shape[-1] == size
    assert feature_transformer.input_dtype == torch.float32


@pytest.mark.parametrize("virtual_batch_size", [None, 7])
@pytest.mark.parametrize("output_size", [10, 12])
@pytest.mark.parametrize("size", [4, 8])
@pytest.mark.parametrize("input_size", [2, 6])
@pytest.mark.parametrize("entmax_mode", [None, "entmax15", "adaptive", "constant"])
@pytest.mark.parametrize("batch_size", [1, 16])
def test_attentive_transformer(
    entmax_mode: Optional[str],
    input_size: int,
    size: int,
    output_size: int,
    virtual_batch_size: Optional[int],
    batch_size: int,
) -> None:
    # setup synthetic tensors
    torch.manual_seed(RANDOM_SEED)
    input_tensor = torch.randn([batch_size, input_size], dtype=torch.float32)
    prior_scales = torch.ones([batch_size, input_size])

    # setup required transformers for test
    feature_transformer = FeatureTransformer(input_size, size + output_size, bn_virtual_bs=virtual_batch_size)
    attentive_transformer = AttentiveTransformer(
        size, input_size, bn_virtual_bs=virtual_batch_size, entmax_mode=entmax_mode
    )

    # process synthetic tensor through transformers
    x = feature_transformer(input_tensor)
    output_tensor = attentive_transformer(x[:, output_size:], prior_scales)

    # check for expected shape and properties
    assert isinstance(output_tensor, torch.Tensor)
    assert output_tensor.shape == (batch_size, input_size)

    assert attentive_transformer.input_shape[-1] == size
    assert attentive_transformer.output_shape[-1] == input_size
    assert attentive_transformer.input_dtype == torch.float32
    if entmax_mode == "adaptive":
        assert isinstance(attentive_transformer.trainable_alpha, torch.Tensor)

    # TODO:  Need variant of assert_model_parameters_updated() to account for the two step calling sequence
    #        of AttentiveTransformer


@pytest.mark.parametrize("virtual_batch_size", [None, 7])
@pytest.mark.parametrize("size", [2, 4, 8])
@pytest.mark.parametrize("output_size", [2, 4, 12])
@pytest.mark.parametrize("input_size", [2])
@pytest.mark.parametrize("entmax_mode", [None, "entmax15", "adaptive", "constant"])
@pytest.mark.parametrize("batch_size", [1, 16])
def test_tabnet(
    entmax_mode: Optional[str],
    input_size: int,
    output_size: int,
    size: int,
    virtual_batch_size: Optional[int],
    batch_size: int,
) -> None:
    # setup synthetic tensor
    torch.manual_seed(RANDOM_SEED)
    input_tensor = torch.randn([batch_size, input_size], dtype=torch.float32)

    tabnet = TabNet(
        input_size, size, output_size, num_steps=3, num_total_blocks=4, num_shared_blocks=2, entmax_mode=entmax_mode
    )

    output = tabnet(input_tensor)

    # check for expected shape and properties
    assert isinstance(output, tuple)
    assert output[0].shape == (batch_size, output_size)

    assert tabnet.input_shape[-1] == input_size
    assert tabnet.output_shape[-1] == output_size
    assert tabnet.input_dtype == torch.float32

    # check for parameter updates
    target = torch.randn([batch_size, 1])
    fpc, tpc, upc, not_updated = check_module_parameters_updated(tabnet, (input_tensor,), target)

    if batch_size == 1:
        # for single record batches, batchnorm layer is bypassed, only a subset of parameters are updated
        assert upc == 15, (
            f"Updated parameter count not expected value. Parameters not updated: {not_updated}"
            f"\nModule structure:\n{tabnet}"
        )
    else:
        # update count should equal trainable number of parameters
        assert tpc == upc, (
            f"All parameter not updated. Parameters not updated: {not_updated}" f"\nModule structure:\n{tabnet}"
        )



import pandas as pd

from ludwig.api import LudwigModel
from ludwig.constants import COLUMN, INPUT_FEATURES, PREPROCESSING, SPLIT
from ludwig.data.split import get_splitter


def filter_cols(df, cols):
    cols = {c.lower() for c in cols}
    retain_cols = [c for c in df.columns if c.lower() in cols]
    return df[retain_cols]


def prepare_data(model: LudwigModel, inputs_df: pd.DataFrame, sample_df: pd.DataFrame, target: str):
    config = model.config
    feature_cols = [feature[COLUMN] for feature in config[INPUT_FEATURES]]
    if SPLIT in config.get(PREPROCESSING, {}):
        # Keep columns required for Ludwig preprocessing
        splitter = get_splitter(**config[PREPROCESSING][SPLIT])
        feature_cols += splitter.required_columns
    target_feature_name = get_feature_name(model, target)

    inputs_df = filter_cols(inputs_df, feature_cols)
    if sample_df is not None:
        sample_df = filter_cols(sample_df, feature_cols)

    return inputs_df, sample_df, feature_cols, target_feature_name


def get_pred_col(preds, target):
    t = target.lower()
    for c in preds.keys():
        if c.lower() == t:
            if "probabilities" in preds[c]:
                return preds[c]["probabilities"]
            else:
                return preds[c]["predictions"]
    raise ValueError(f"Unable to find target column {t} in {preds.keys()}")


def get_feature_name(model: LudwigModel, target: str) -> str:
    t = target.lower()
    for c in model.training_set_metadata.keys():
        if c.lower() == t:
            return c
    raise ValueError(f"Unable to find target column {t} in {model.training_set_metadata.keys()}")



from typing import Any, Dict, List, Optional, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import common_fields
from ludwig.schema import utils as schema_utils
from ludwig.schema.combiners.base import BaseCombinerConfig
from ludwig.schema.metadata import COMBINER_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class ConcatCombinerConfig(BaseCombinerConfig):
    """Parameters for concat combiner."""

    @staticmethod
    def module_name():
        return "ConcatCombiner"

    type: str = schema_utils.ProtectedString(
        "concat",
        description=COMBINER_METADATA["ConcatCombiner"]["type"].long_description,
    )

    dropout: float = common_fields.DropoutField()

    activation: str = schema_utils.ActivationOptions(default="relu")

    flatten_inputs: bool = schema_utils.Boolean(
        default=False,
        description="Whether to flatten input tensors to a vector.",
        parameter_metadata=COMBINER_METADATA["ConcatCombiner"]["flatten_inputs"],
    )

    residual: bool = common_fields.ResidualField()

    use_bias: bool = schema_utils.Boolean(
        default=True,
        description="Whether the layer uses a bias vector.",
        parameter_metadata=COMBINER_METADATA["ConcatCombiner"]["use_bias"],
    )

    bias_initializer: Union[str, Dict] = common_fields.BiasInitializerField()

    weights_initializer: Union[str, Dict] = common_fields.WeightsInitializerField()

    num_fc_layers: int = common_fields.NumFCLayersField()

    output_size: int = schema_utils.PositiveInteger(
        default=256,
        description="Output size of a fully connected layer.",
        parameter_metadata=COMBINER_METADATA["ConcatCombiner"]["output_size"],
    )

    norm: Optional[str] = common_fields.NormField()

    norm_params: Optional[dict] = common_fields.NormParamsField()

    fc_layers: Optional[List[Dict[str, Any]]] = common_fields.FCLayersField()



# TODO(travis): figure out why we need these imports to avoid circular import error
from ludwig.schema.combiners.utils import get_combiner_jsonschema  # noqa
from ludwig.schema.features.utils import get_input_feature_jsonschema, get_output_feature_jsonschema  # noqa
from ludwig.schema.hyperopt import get_hyperopt_jsonschema  # noqa
from ludwig.schema.preprocessing import get_preprocessing_jsonschema  # noqa
from ludwig.schema.trainer import get_model_type_jsonschema, get_trainer_jsonschema  # noqa



import json
import os
import tempfile
from typing import Any, Dict, Union

import numpy as np
import pandas as pd
from starlette.datastructures import UploadFile
from starlette.responses import JSONResponse

from ludwig.utils.data_utils import NumpyEncoder


def serialize_payload(data_source: Union[pd.DataFrame, pd.Series]) -> tuple:
    """
    Generates two dictionaries to be sent via REST API for Ludwig prediction
    service.
    First dictionary created is payload_dict. Keys found in payload_dict:
    raw_data: this is json string created by pandas to_json() method
    source_type: indicates if the data_source is either a pandas dataframe or
        pandas series.  This is needed to know how to rebuild the structure.
    ndarray_dtype:  this is a dictionary where each entry is for any ndarray
        data found in the data_source.  This could be an empty dictioinary if no
        ndarray objects are present in data_source. Key for this dictionary is
        column name if data_source is dataframe or index name if data_source is
        series.  The value portion of the dictionary is the dtype of the
        ndarray.  This value is used to set the correct dtype when rebuilding
        the entry.

    Second dictionary created is called payload_files, this contains information
    and content for files to be sent to the server.  NOTE: if no files are to be
    sent, this will be an empty dictionary.
    Entries in this dictionary:
    Key: file path string for file to be sent to server
    Value: tuple(file path string, byte encoded file content,
                 'application/octet-stream')

    Args:
        data_source: input features to be sent to Ludwig server

    Returns: tuple(payload_dict, payload_files)

    """
    payload_dict = {}
    payload_dict["ndarray_dtype"] = {}
    payload_files = {}
    if isinstance(data_source, pd.DataFrame):
        payload_dict["raw_data"] = data_source.to_json(orient="columns")
        payload_dict["source_type"] = "dataframe"
        for col in data_source.columns:
            if isinstance(data_source[col].iloc[0], np.ndarray):
                # if we have any ndarray columns, record dtype
                payload_dict["ndarray_dtype"][col] = str(data_source[col].iloc[0].dtype)
            elif isinstance(data_source[col].iloc[0], str) and os.path.exists(data_source[col].iloc[0]):
                # if we have file path feature, prepare file for transport
                for v in data_source[col]:
                    payload_files[v] = (v, open(v, "rb"), "application/octet-stream")
    elif isinstance(data_source, pd.Series):
        payload_dict["raw_data"] = data_source.to_json(orient="index")
        payload_dict["source_type"] = "series"
        for col in data_source.index:
            if isinstance(data_source[col], np.ndarray):
                # for ndarrays record dtype for reconstruction
                payload_dict["ndarray_dtype"][col] = str(data_source[col].dtype)
            elif isinstance(data_source[col], str) and os.path.exists(data_source[col]):
                # if we have file path feature, prepare file for transport
                v = data_source[col]
                payload_files[v] = (v, open(v, "rb"), "application/octet-stream")
    else:
        ValueError(
            '"data_source" must be either a pandas DataFrame or Series, '
            "format found to be {}".format(type(data_source))
        )

    return payload_dict, payload_files


def _write_file(v, files):
    # Convert UploadFile to a NamedTemporaryFile to ensure it's on the disk
    suffix = os.path.splitext(v.filename)[1]
    named_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
    files.append(named_file)
    named_file.write(v.file.read())
    named_file.close()
    return named_file.name


def deserialize_payload(json_string: str) -> pd.DataFrame:
    """This function performs the inverse of the serialize_payload function and rebuilds the object represented in
    json_string to a pandas DataFrame.

    Args:
        json_string: representing object to be rebuilt.

    Returns: pandas.DataFrame
    """
    payload_dict = json.loads(json_string)

    # extract raw data from json string
    raw_data_dict = json.loads(payload_dict["raw_data"])
    # rebuild based on original data source
    if payload_dict["source_type"] == "dataframe":
        # reconstitute the pandas dataframe
        df = pd.DataFrame.from_dict(raw_data_dict, orient="columns")
    elif payload_dict["source_type"] == "series":
        # reconstitute series into single row dataframe
        df = pd.DataFrame(pd.Series(raw_data_dict)).T
    else:
        ValueError(
            'Unknown "source_type" found.  Valid values are "dataframe" or '
            '"series".  Instead found {}'.format(payload_dict["source_type"])
        )

    # if source has ndarrays, rebuild those from list and set
    # original dtype.
    if payload_dict["ndarray_dtype"]:
        # yes, now covert list representation to ndarray representation
        for col in payload_dict["ndarray_dtype"]:
            dtype = payload_dict["ndarray_dtype"][col]
            df[col] = df[col].apply(lambda x: np.array(x).astype(dtype))

    return df


def deserialize_request(form) -> tuple:
    """This function will deserialize the REST API request packet to create a pandas dataframe that is input to the
    Ludwig predict method and a list of files that will be cleaned up at the end of processing.

    Args:
        form: REST API provide form data

    Returns: tuple(pandas.DataFrame, list of temporary files to clean up)
    """
    files = []
    file_index = {}
    for k, v in form.multi_items():
        if type(v) == UploadFile:
            file_index[v.filename] = _write_file(v, files)

    # reconstruct the dataframe
    df = deserialize_payload(form["payload"])

    # insert files paths of the temporary files in place of the original
    # file paths specified by the user.
    # pd.DataFrame.replace() method is used to replace file path string
    # specified by the user context with the file path string where a
    # temporary file containing the same content.
    # parameters for replace() method:
    #   to_replace: list of file path strings that the user provided
    #   value: list of temporary files created for each input file
    #
    # IMPORTANT: There is a one-to-one correspondence of the to_replace list
    # and the value list. Each list must be the same size.
    df.replace(to_replace=list(file_index.keys()), value=list(file_index.values()), inplace=True)

    return df, files


class NumpyJSONResponse(JSONResponse):
    def render(self, content: Dict[str, Any]) -> str:
        """Override the default JSONResponse behavior to encode numpy arrays.

        Args:
            content: JSON object to be serialized.

        Returns: str
        """
        return json.dumps(
            content, ensure_ascii=False, allow_nan=False, indent=None, separators=(",", ":"), cls=NumpyEncoder
        ).encode("utf-8")



from typing import Dict

import pytest
import torch

from ludwig.constants import ENCODER, TYPE
from ludwig.features.timeseries_feature import TimeseriesInputFeature
from ludwig.schema.features.timeseries_feature import TimeseriesInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from ludwig.utils.torch_utils import get_torch_device

SEQ_SIZE = 2
TIMESERIES_W_SIZE = 1
MAX_LEN = 7
EMBEDDING_SIZE = 5
DEVICE = get_torch_device()


@pytest.fixture(scope="module")
def timeseries_config():
    return {
        "name": "timeseries_12",
        "type": "timeseries",
        "encoder": {
            "max_len": MAX_LEN,
            "embedding_size": EMBEDDING_SIZE,
            "max_sequence_length": SEQ_SIZE,
            "output_size": 8,
            "state_size": 8,
            "num_filters": 8,
            "hidden_size": 8,
        },
    }


@pytest.mark.parametrize("encoder", ["rnn", "stacked_cnn", "parallel_cnn"])
def test_timeseries_input_feature(timeseries_config: Dict, encoder: str) -> None:
    timeseries_config[ENCODER][TYPE] = encoder

    timeseries_config, _ = load_config_with_kwargs(TimeseriesInputFeatureConfig, timeseries_config)
    timeseries_input_feature = TimeseriesInputFeature(timeseries_config).to(DEVICE)
    timeseries_tensor = torch.randn([SEQ_SIZE, TIMESERIES_W_SIZE], dtype=torch.float32).to(DEVICE)
    encoder_output = timeseries_input_feature(timeseries_tensor)
    assert encoder_output["encoder_output"].shape[1:] == timeseries_input_feature.output_shape
